""" core LazyMixin functionality 

be very cautious of referencing self.<attribute> when there is no 'lazy' prefix to it

for this reason auxiliary classes have been moved to core_assist.py

"""

import pdb
import os

# pylint: disable=unused-wildcard-import,wildcard-import,unused-import
from .common import *

#######################################################
# Typing
#######################################################
from typing import (
    Optional,
    # TYPE_CHECKING,
    Any,
    Union,
)

from lazy_regression_tests._baseutils import ppp, Dummy

from traceback import print_exc as xp


# pylint: enable=unused-wildcard-import,wildcard-import,unused-import

from .validators import build_validators_for_class
from .filters import build_filters_for_class


#######################################################


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb


class LazyChecker:

    extension = "?"

    def __repr__(self):
        return "%s[%s]" % (self.__class__.__name__, self.extension)

    def __init__(
        self,
        extension: str,
        rawfiltermgr=None,
        textfiltermgr=None,
        # write_exp_on_ioerror: bool = True,
    ):
        self.lazy_extension = self.extension = extension

        # self.write_exp_on_ioerror = write_exp_on_ioerror
        self.rawfiltermgr = rawfiltermgr
        self.textfiltermgr = textfiltermgr

        self.filterhash = None

        self.reg_callbacks = {}

    def debug(self, ix, lines):
        print("\n".join(lines[ix - 5 : +ix + 5]))

    def prep(self, tmp, data):
        return data

    def to_text(self, tmp, data):
        return str(data)

    def filter_raw(self, tmp, data):
        return self.rawfiltermgr.filter(self, tmp, data)

    def filter_text(self, tmp, data):
        return self.textfiltermgr.filter(self, tmp, data).strip()

    def format(self, tmp, data, fnp_raw):
        try:
            # used to only format once
            if (
                isinstance(data, str)
                and self.filterhash
                and self.filterhash == hash(data)
            ):
                return data

            data = self.prep(tmp, data)

            if fnp_raw:
                str_data = self.to_text(tmp, data)

                dirname = os.path.dirname(fnp_raw)
                if not os.path.isdir(dirname):
                    os.makedirs(dirname)

                with open(fnp_raw, "w") as fo:
                    fo.write(str_data)

            data = self.filter_raw(tmp, data)
            str_data = self.to_text(tmp, data)
            str_data = self.filter_text(tmp, str_data)

            self.filterhash = hash(str_data)

            return str_data
        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                ppp(self, self)
                pdb.set_trace()
            raise


class ExecInfo:
    def __init__(self):
        self.ioerror_exp = self.ioerror_got = None


class LazyTemp(object):
    def __repr__(self):
        return "LazyTemp[id=%s]" % (id(self))

    def __init__(self, control, env, testee):
        self.control = control
        self.fnp_exp = self.fnp_got = None
        self.env = env.copy()
        self.message = ""
        self.filtered = Dummy()
        self.testee = testee

        self.execinfo = ExecInfo()

    def add_filtered(self, name, value, scalar):
        """ each filter saves what it finds here """
        try:
            if scalar:
                setattr(self.filtered, name, value)
            else:
                li = getattr(self.filtered, name, None)
                if li is None:
                    li = []
                    setattr(self.filtered, name, li)

                li.append(value)

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class MediatedEnvironDict(dict):
    def __init__(self, acquired=False, **kwds):
        """
        """
        self.acquired = acquired
        super(MediatedEnvironDict, self).__init__(**kwds)

    def acquire(self, rootname):
        if self.acquired:
            return

        len_root = len(rootname)

        for k, value in os.environ.items():
            if k.startswith(rootname):
                k2 = k[len_root:]
                self[k2] = value

        self.acquired = True


class _Control(object):
    """unifies environment and function arguments
       to determine handlers for IOError and AssertionError
       save in the LazyTemp results object as well.
    """

    def __init__(
        self, lazy: "LazyMixin", env: MediatedEnvironDict, options: "LazyCheckerOptions"
    ):

        self.lazy = lazy
        self.env = env
        self.options = options

    _save_raw = undefined

    @property
    def save_raw(self):
        """ this depend on 2 environmental settings:
            :param lzrt_save_raw_received_data : if it it's set we can do it
            :param lzrt_template_dirname_report : where to put it
        """

        if self._save_raw is undefined:
            res = bool(self.env.get("template_dirname_report"))
            self._save_raw = bool(res and self.env.get("save_raw_received_data"))
        return self._save_raw

    _directive = undefined

    @property
    def directive(self) -> str:
        """
        This value is obtained from the environment variable, type `lzrt_directive`

        To set it on the command line, use standard Unix procedures, i.e. to use baseline behavior:

        `lzrt_directive=baseline pytest ....`
        `lzrt_directive=baseline python ....`

        The directive defines 3 mutually exclusive ways for lazy tests to behave, with an additional
        behavior when it is unset i.e. the default behavior:

        :baseline: this will take any received data, write it to expectations file and return a success
        :nodiff  : a `assertTrue(exp==got)` comparison will be run, rather than an `assertEqual`
        :skip    : lazytesting will not be performed at at all and success is automatic since no asserts
        are run

        default behavior, when unset:

        - `exp` file is loaded and compared to `got` data
        - if the `exp` file is missing: 
            - the `got` data is written to both the `exp` and `got` file
            - return success
        """

        if self._directive is undefined:
            res = self._directive = self.env.get("directive", "").strip().lower()
            assert res in (
                OPT_DIRECTIVE_SKIP,
                OPT_DIRECTIVE_BASELINE,
                OPT_DIRECTIVE_NODIFF,
                "",
            )
        return self._directive

    def skip(self):
        return self.directive == OPT_DIRECTIVE_SKIP

    def baseline(self):
        return self.directive == OPT_DIRECTIVE_BASELINE

    def nodiff(self):
        return self.directive == OPT_DIRECTIVE_NODIFF


#######################################################
#
#######################################################


class _LazyMeta(type):
    def __init__(cls, name, bases, attrs, **kwargs):
        """ 
        intercepting the newly created class allows stacking of the 
        ancestral validators and formatters in reverse mro order
        i.e. top of ancestors go first
        """

        try:

            classname = cls.__name__

            # we want to build out the validators by running the basic ancestral ones first
            # ex:  check status=200 and content_type = html before checking <title>
            li_bases2current = list(reversed(cls.mro()))

            li_ancestor_filter = []

            for basecls in li_bases2current:

                cls_filters = getattr(basecls, "cls_filters", {})

                if cls_filters:
                    li_ancestor_filter.append(cls_filters)

            cls.cls_validators = build_validators_for_class(cls, li_bases2current)

            cls.cls_filters = build_filters_for_class(cls, li_ancestor_filter)

            super().__init__(name, bases, attrs)

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise



import os
import pdb
import tempfile
import datetime
import random
import string

from traceback import print_exc as xp


from lazy_regression_tests.utils import ppp, getpath, UnavailableLibrary

#################################################################
# Depencies for test
#################################################################


class Foo:
    pass


dep_message = "👉 This is only requiry for lazy self-testing on http contents, not for normal use. 👈"


try:
    import responses
except (ImportError,) as e:
    # this will throw an InvalidConfigurationError on any access to bs
    # telling you to install BeautifulSoup

    responses = UnavailableLibrary(
        name=Foo.__module__, missing="responses", message=dep_message
    )


try:
    import requests
except (ImportError,) as e:
    # this will throw an InvalidConfigurationError on any access to bs
    # telling you to install BeautifulSoup
    requests = UnavailableLibrary(
        name=Foo.__module__, missing="requests", message=dep_message
    )

try:
    import jinja2
except (ImportError,) as e:
    # this will throw an InvalidConfigurationError on any access to bs
    # telling you to install BeautifulSoup
    jinja2 = UnavailableLibrary(
        name=Foo.__module__, missing="jinja2", message=dep_message
    )


try:
    import unittest.mock as mock
except (ImportError,) as ei:
    import mock  # python 2?


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb


from lazy_regression_tests.http_validators import ResponseHTML


def get_mock_env(seed={}):
    """
        this is used to mock the environment variables and 

        also determines whether temp files are:
        - used     : useful for self-testing the lazy-test itself
        - not used : pretty much required for actual lazy-test usage

        Note the self-caching 
    """

    try:

        # to use temp files or not?
        USE_TEMP_FILE = not (os.getenv("lzrt_TESTING_NO_TEMP_FILE"))

        if USE_TEMP_FILE:

            di_mock_env = getattr(get_mock_env, "di_mock_env", None)
            if di_mock_env:
                res = di_mock_env.copy()
                res.update(**seed)
                return res

            dirtemp = tempfile.mkdtemp(prefix="lazy_regression_tests_")

            lzrt_default_t_basename = "%(filename)s %(classname)s %(_testMethodName)s %(lazy_basename_extras)s %(suffix)s %(extension)s"

            lzrt_template_dirname = os.path.join(dirtemp, "out")
            lzrt_template_dirname_got = os.path.join(dirtemp, "got")
            lzrt_template_dirname_exp = os.path.join(dirtemp, "exp")
            lzrt_template_dirname_report = os.path.join(dirtemp, "report")

            lzrt_template_basename = lzrt_default_t_basename

            res = get_mock_env.di_mock_env = dict(
                lzrt_template_dirname=lzrt_template_dirname,
                lzrt_template_dirname_got=lzrt_template_dirname_got,
                lzrt_template_dirname_exp=lzrt_template_dirname_exp,
                lzrt_template_basename=lzrt_template_basename,
                lzrt_template_dirname_report=lzrt_template_dirname_report,
            )

            res = res.copy()
            res.update(**seed)

            return res
        else:
            res = os.environ.copy()
            res.update(**seed)
            return res

    # pragma: no cover pylint: disable=unused-variable
    except (Exception,) as e:
        if cpdb():
            pdb.set_trace()
        raise


choice_csrf = string.ascii_letters + string.digits


def get_fake_response_from_html(
    html, url=None, status=200, content_type="text/html; charset=utf8"
):
    try:

        url = url or "http://example.com/"

        @responses.activate
        def faker(html):
            # responses.add(responses.GET, 'http://twitter.com/api/1/foobar',
            #               json={'error': 'not found'}, status=404)

            responses.add(
                responses.GET, url, body=html, status=status, content_type=content_type
            )

            resp = requests.get(url)
            return resp

        response = faker(html)
        assert response
        return response

    # pragma: no cover pylint: disable=unused-variable
    except (Exception,) as e:
        if cpdb():
            pdb.set_trace()
        raise


def get_fake_response_from_template(
    testee, data={}, url=None, status=200, content_type="text/html; charset=utf8"
):
    """ return a pretend HttpResponse"""
    try:

        url = url or "http://example.com/"

        data = testee.get_data(seed=data)

        tmpl = jinja2.Template(testee.template)
        text = tmpl.render(**data)

        response = get_fake_response_from_html(html=text, url=url, status=status)

        assert response

        return response

    # pragma: no cover pylint: disable=unused-variable
    except (Exception,) as e:
        if cpdb():
            pdb.set_trace()
        raise


exp_fail_padder = "✅" * 40


def show_expected_fail(exc):
    """ show the error details on expected failures"""
    print(f"{exp_fail_padder}{exc}{exp_fail_padder}")


class Helper:

    fail_on_validate = None
    fail_on_change = None

    def simulate_previous_pass(self):
        """ this simulates a first run"""

        http_response = self.get_raw_response()
        response = ResponseHTML(http_response)
        self.assert_exp(response.content, self.extension)

    def get_raw_response(self, url=None, data={}):
        try:

            data = self.get_data(seed=data)

            tmpl = jinja2.Template(self.template)
            html = tmpl.render(**data)

            response = get_fake_response_from_html(html, url=url)
            return response

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def get_data(self, seed={}):
        data = dict(vars(self.__class__))
        data.update(timestamp=datetime.datetime.now())

        csrf = "".join([random.choice(choice_csrf) for ix in range(0, 20)])

        data.update(csrf=csrf)
        data.update(**vars(self))
        data.update(**getattr(self, "data", {}))

        data.update(**seed)

        return data

    def get_response(self, url=None, data={}):
        try:

            data = data or self.get_data(seed=data)

            tmpl = Template(self.template)
            text = tmpl.render(**data)
            return ResponseHTML(HttpResponse(text))

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class HelperHTML(Helper):
    def get_response(self, url=None, data={}):
        try:

            http_response = self.get_raw_response(url=url, data=data)
            return ResponseHTML(http_response)

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class CheckitMixin:

    di_mock_env = get_mock_env()

    @mock.patch.dict(os.environ, di_mock_env)
    def test_it(self):
        """get response"""
        try:

            # first pass
            response = self.get_response()
            self.assert_exp(response.content, self.extension)

            response = self.get_response()

            if self.fail_on_validate:
                try:
                    self.check_expectations(response=response)
                # pragma: no cover pylint: disable=unused-variable
                except (AssertionError,) as e:
                    self.assertTrue(str(self.fail_on_validate) in str(e))
                else:
                    self.fail(
                        "should have failed validation with %s"
                        % (self.fail_on_validate)
                    )

            else:
                self.check_expectations(response=response)

            if self.fail_on_change:
                try:
                    tmp = self.assert_exp(response.content, self.extension)
                # pragma: no cover pylint: disable=unused-variable
                except (AssertionError,) as e:
                    self.assertTrue(str(self.fail_on_change) in str(e))
                else:
                    self.fail(
                        "should have failed change checks with %s"
                        % (self.fail_on_change)
                    )

            else:
                tmp = self.assert_exp(response.content, self.extension)

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


def main():
    @responses.activate
    def test_simple():
        # responses.add(responses.GET, 'http://twitter.com/api/1/foobar',
        #               json={'error': 'not found'}, status=404)

        responses.add(
            responses.GET,
            "http://twitter.com/api/1/foobar",
            body="<div>coucou</div>",
            status=200,
        )

        resp = requests.get("http://twitter.com/api/1/foobar")

        return resp

    response = test_simple()
    ppp(response, "response")

    pdb.set_trace()


if __name__ == "__main__":
    main()



import pdb
import sys
import re
import os

from operator import attrgetter

import logging

logger = logging.getLogger(__name__)

from lazy_regression_tests._baseutils import (
    debugObject,
    ppp,
    Dummy,
    getpath,
    InvalidConfigurationException,
)

from traceback import print_exc as xp

from .common import *


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb


#######################################################
# Typing
#######################################################
from typing import (
    Optional,
    # TYPE_CHECKING,
    Any,
    Union,
)


from lazy_regression_tests.utils import nested_dict_get, first, fill_template, ppp


#######################################################
# constants
#######################################################

verbose = "-v" in sys.argv

undefined = NotImplemented

# https://stackoverflow.com/questions/6102019/type-of-compiled-regex-object-in-python
# yeah, it is what it is, so be it...
type_regex_hack = re.compile("").__class__


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb


class DoNotCheck:
    """use this to flag attributes that should not get checked """


class AutoExp:
    """ this class can be used to specify exp in 2 ways 

        at class level, `directive.exp is AutoExp` means that 

            directive.exp = getattr(testee, directive.name)

        at instance level `isistance(directive.exp, Autoexp)` will:

            for path in directive.exp.paths:
                try:

    """

    def __repr__(self):
        return "%s[%s]" % (self.__class__.__name__, getattr(self, "paths", ""))

    required = True

    @classmethod
    def get_exp_by_name(cls, testee, name):
        try:
            return getpath(testee, name)

        # pragma: no cover pylint: disable=unused-variable
        except (AttributeError,) as e:
            raise InvalidConfigurationException(
                "Attribute/Key %s not found on %s but expected via AutoExp.  either disable the validation or add this attribute"
                % (name, testee),
                ori=e,
            )

        # pragma: no cover pylint: disable=unused-variable

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def get_exp(self, testee):
        try:

            for path in self.paths:
                try:
                    return getpath(testee, path)
                # pragma: no cover pylint: disable=unused-variable
                except (AttributeError,) as e:
                    pass
            else:
                # raise AttributeError(self.paths)
                raise InvalidConfigurationException(
                    "Attribute/Key %s not found on %s but expected via AutoExp.  either disable the validation or add this attribute"
                    % (self.paths, testee)
                )
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def __init__(self, paths: Union[list, tuple, str]):
        try:
            if isinstance(paths, tuple):

                self.paths = paths

            elif isinstance(paths, list):
                self.paths = tuple(list)

            elif isinstance(paths, str):
                self.paths = (paths,)

            else:
                TypeError("%s.__init__:paths should be tuple, list or str of str")

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class OptAutoExp(AutoExp):
    required = False


#######################################################
# Validators
#######################################################


class Validator:
    """ base validator functionality 
    - get source during check_expectations, testee is used if sourcename is empty
    - retrieve got from get_value from the source
    - call a default assert exp == got
    """

    attrname: str
    sourcename: str

    def __repr__(self):
        return "%s[selector=%s]" % (
            self.__class__.__name__,
            getattr(self, "selector", "?"),
        )

    def __init__(
        self, selector, sourcename: Optional[str] = None, scalar=True, cargo=None
    ):
        try:
            self.selector = selector
            if sourcename:
                self.sourcename = sourcename
            self.scalar = scalar
            self.cargo = cargo

            assert isinstance(sourcename, str) or sourcename is None
        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise

    def get_source(self, testee, **sources):

        if self.sourcename is None:
            source_ = testee

        else:
            if "." in self.sourcename:
                try:
                    source_ = getpath(sources, self.sourcename)
                # pragma: no cover pylint: disable=unused-variable
                except (AttributeError,) as e:
                    raise ValueError(
                        "%s expects source:%s: which is not in the provided %s"
                        % (self, self.sourcename, ",".join(sources.keys()))
                    )
            else:
                try:
                    source_ = sources[self.sourcename]
                # pragma: no cover pylint: disable=unused-variable
                except (KeyError,) as e:
                    raise ValueError(
                        "%s expects source:%s: which is not in the provided %s"
                        % (self, self.sourcename, ",".join(sources.keys()))
                    )

        if source_ is None:
            raise ValueError(
                "you need to either pass in a dict in `source` or %s needs to have `%s` set"
                % (testee, self.sourcename)
            )
        return source_

    def get_value(self, source):
        if self.selector is None:
            return source
        else:
            res = source.get(self.selector)
        if rpdb() and res is None:  # pragma: no cover
            pdb.set_trace()
        return res

    def check(self, name: str, testee: "LazyMixin", exp: Any, sources: dict):
        """

        """

        try:

            source_ = self.get_source(testee, **sources)
            try:
                got = self.get_value(source_)
            # pragma: no cover pylint: disable=unused-variable
            except (KeyError, AttributeError) as e:
                logger.warning(
                    "%s.validation error: could not access data in source %s"
                    % (self, e)
                )
                # pdb.set_trace()
                got = undefined

            t_message = getattr(self, "t_message", None)
            message = (
                fill_template(t_message, locals(), testee, self) if t_message else None
            )

            if isinstance(exp, type_regex_hack):
                self.test_regex(testee, exp, got, message, name=name)

            elif not callable(exp):
                self.test(testee, exp, got, message, name=name)
                return got
            else:
                exp(testee=testee, got=got, validator=self)
                return got
        except (AssertionError,) as e:  # pragma: no cover
            raise
        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise

    def test_regex(self, testee, exp, got, message, name=None):
        try:

            if message is None:
                message = fill_template(
                    "%(name)s pattern:%(pattern)s:does not match:%(got)s:got",
                    locals(),
                    exp,
                    self,
                    {"name": name or self},
                )

            if not isinstance(got, (list, set)):
                testee.assertTrue(exp.search(str(got)), message)
            else:
                for igot in got:
                    hit = exp.search(str(igot))
                    if hit is None:
                        message = f"validation:{name or ''} failure: data:'{str(igot)}': does not match regex:'{exp.pattern}':"
                        testee.fail(message)

        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise

    def test(self, testee, exp, got, message, name=None):
        try:

            if exp is undefined:
                raise ValueError("exp is undefined")

            if message is None:
                message = fill_template(
                    "%(name)s exp:%(exp)s<>%(got)s:got",
                    locals(),
                    self,
                    {"name": name or self},
                )

            # correcting for unspecified scalars
            if not isinstance(exp, list) and isinstance(got, list) and len(got) == 1:
                got = first(got)

            testee.assertEqual(exp, got, message)
        # except (AssertionError,) as e:  # pragma: no cover
        #     raise

        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise


class AttributeValidator(Validator):
    def __init__(self, selector, sourcename):
        super(AttributeValidator, self).__init__(selector, self.sourcename)
        self.f_getter = attrgetter(selector)

    def get_value(self, source_):
        try:
            res = self.f_getter(source_)
            if rpdb() and res is None:  # pragma: no cover
                pdb.set_trace()
            return res
        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise


class FullyQualifiedNamesValidator(Validator):
    def __init__(self):
        Validator.__init__(self, self.selector, self.sourcename)


class DictValidator(Validator):

    sourcename = "data"

    def get_value(self, source_):
        """ get the value from  """
        try:
            res = nested_dict_get(source_, self.selector)
            return res
        except (KeyError,) as e:  # pragma: no cover
            raise
        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise


class AttrNamedDictValidator(DictValidator):
    def __init__(self, selector):
        super(AttrNamedDictValidator, self).__init__(
            selector, scalar=True, sourcename=self.sourcename
        )


class MixinExpInGot:
    def test(self, testee, exp, got, message, name=None):

        if exp is undefined:
            raise ValueError("exp is undefined")
        try:

            if message is None:
                try:
                    message = "%s : exp:%s: is not in got:%s:" % (
                        name or self,
                        str(exp),
                        str(got),
                    )
                # pragma: no cover pylint: disable=unused-variable, broad-except
                except (Exception,) as ignore_:
                    # well, that didn't work
                    pass

            testee.assertTrue(str(exp) in str(got), message)

        # pragma: no cover pylint: disable=unused-variable
        except (AssertionError,) as e:
            pdb.set_trace()
            raise
        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise


class DirectValidator(AttributeValidator):
    """ this is intended to check attributes/data on unittest.Testcase instance """

    sourcename = None

    def __init__(self, selector=None):

        selector = selector or getattr(self, "selector", None)
        if selector is None:
            raise ValueError(
                "selector was not passed in and is not pre-set on class %s"
                % (self.__class__.__name__)
            )

        super(NamedTesteeAttributeValidator, self).__init__(
            selector, sourcename=self.sourcename
        )


NamedTesteeAttributeValidator = DirectValidator


#######################################################
# Managers
#######################################################


class ValidationDirective:
    def __repr__(self):

        try:

            exp = getattr(self, "exp", undefined)
            if exp is undefined:
                exp = "undefined"
            elif isinstance(exp, type):
                exp = exp.__name__
            else:
                exp = str(exp)

            return f"{self.__class__.__name__:20.20}:{self.name:30.30} active:{self.active} exp:{exp:30.30}"

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            return self.__class__.__name__

    def __init__(
        self,
        name: str,
        validator: Optional[Validator] = None,
        exp=None,
        active: Optional[bool] = None,
    ):
        try:
            self.name = name
            self.validator = validator
            self.exp = exp

            # if we know exp and we have a validator
            # assume active unless it has been set to True or False
            if active is None and exp and validator:
                active = True

            assert validator is None or isinstance(validator, Validator)

            self.active = active

        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if 1 or cpdb():
                pdb.set_trace()
            raise

    def copy(self):
        return self.__class__(self.name, self.validator, self.exp, self.active)


class ValidationManager:
    def __repr__(self):

        extra = f" for {self.testee}"

        if verbose:

            lines = ["", "  validators:"]
            lines += ["    %s" % str(val) for val in self.validators.values()]
            extra += "\n".join(lines)

        return "%s%s" % (self.__class__.__name__, extra)

    def __init__(self, testee, *validator_managers):

        self.validators = {}
        self.testee = testee

        for validatormgr in validator_managers:

            if isinstance(validatormgr, ValidationDirective):
                # remember, you need to copy shared directives because set_expectations modifies them
                self._add_baseline(validatormgr.copy())
            elif isinstance(validatormgr, ValidationManager):
                self.inject(validatormgr)

    def prep_validation(self, testee, names):
        try:

            for name in names:
                directive = self.validators[name]

                if directive.exp in (AutoExp, OptAutoExp):
                    try:
                        directive.exp = directive.exp.get_exp_by_name(testee, name)
                        continue
                    except (AttributeError,) as e:
                        if not directive.exp.required:
                            directive.active = False
                            continue
                        raise

                elif isinstance(directive.exp, AutoExp):
                    try:
                        directive.exp = directive.exp.get_exp(testee)
                        continue
                    except (AttributeError,) as e:
                        if not directive.exp.required:
                            directive.active = False
                            continue
                        raise

        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise

    def _get_names(
        self,
        seen: set,
        lazy_skip_except=None,
        lazy_skip=None,
        lazy_sourced_only: bool = True,
    ):

        try:

            keep_filter = None
            if lazy_skip_except:
                if isinstance(lazy_skip_except, str):
                    lazy_skip_except = [lazy_skip_except]

                if isinstance(lazy_skip_except, (dict, list, set)):

                    def keep_filter(name):
                        return name in lazy_skip_except

                elif isinstance(lazy_skip_except, type_regex_hack):

                    def keep_filter(name):
                        return bool(lazy_skip_except.search(name))

                else:
                    raise TypeError(
                        "lazy_skip_except supports dictionary, list, string, set and regex:.got(%s)"
                        % lazy_skip_except
                    )

            skip_filter = None
            if lazy_skip:
                if isinstance(lazy_skip, str):
                    lazy_skip = [lazy_skip]

                if isinstance(lazy_skip, (dict, list, set)):

                    def skip_filter(name):
                        return name in lazy_skip

                elif isinstance(lazy_skip, type_regex_hack):

                    def skip_filter(name):
                        return bool(lazy_skip.search(name))

                else:
                    raise TypeError(
                        "lazy_skip supports dictionary, list, string, set and regex:.got(%s)"
                        % lazy_skip
                    )

            res = []
            for name in self.validators:
                logname = name
                if name == "content_type":
                    directive = self.validators[name]
                    logname = "%s=%s" % (name, directive.exp)

                if skip_filter and skip_filter(name):
                    seen.add("%s.skipped" % (logname))
                    continue

                if keep_filter and not keep_filter(name):
                    seen.add("%s.skipped" % (logname))
                    continue
                res.append(name)

            return res

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def check_expectations(
        self,
        testee: "unittest.TestCase",
        lazy_skip=None,
        lazy_skip_except=None,
        lazy_sourced_only: bool = True,
        **sources,
    ):
        """
        loops through the validation directives and executives them if active
        errors if exp or validator is missing 
        :parm sources: a dictionary that provides data for each validator's sourcename
        note that a validator can leave sourcename empty which means testee is 
        the source
        :param lazy_sourced_only: use this if not all sources are ready yet

        """

        try:

            seen = set()
            names = self._get_names(
                seen=seen, lazy_skip_except=lazy_skip_except, lazy_skip=lazy_skip
            )

            name = validator = None

            self.prep_validation(testee, names)

            for name in names:
                directive = self.validators[name]

                logname = name
                if name == "content_type":
                    logname = "%s=%s" % (name, directive.exp)

                validator = directive.validator

                if not directive.active or directive.active is undefined:
                    seen.add("%s.inactive" % (logname))
                    continue

                # sometimes, a test is done in several phases and not all sources are ready
                if lazy_sourced_only:
                    # takes into account dotted paths and empty sourcenames which point
                    # to testee
                    sourcename = (getattr(validator, "sourcename", "") or "").split(
                        "."
                    )[0]

                    if sourcename and not sourcename in sources:
                        seen.add("%s.unsourced" % (logname))
                        continue

                exp = directive.exp

                # maybe it's built into the validator?
                if exp is undefined:
                    exp = getattr(directive.validator, "exp", undefined)

                if exp is undefined:
                    # let's see if we can get it
                    exp = getattr(testee, directive.name, undefined)

                if exp is undefined:
                    raise InvalidConfigurationException(
                        "%s has undefined exp" % (directive)
                    )

                seen.add(logname)
                validator.check(name, testee, exp, sources)

        except (AssertionError,) as assert_exc:  # pragma: no cover
            format_assertion_error(self, assert_exc, validator=validator, name=name)
            raise
        # pragma: no cover pylint: disable=unused-variable, broad-except
        except (Exception,) as e:  #
            if cpdb():
                pdb.set_trace()
            raise

    def inject(self, validatormgr):

        for name, directive in validatormgr.validators.items():
            # remember, you need to copy shared directives because set_expectations modifies them
            self._add_baseline(directive.copy())

    def _add_baseline(self, directive, name=None):
        name = name or directive.name

        try:

            existing = self.validators.get(name)
            if existing:

                if directive.active is not None:
                    existing.active = directive.active

                if directive.exp is not undefined:
                    existing.exp = directive.exp

                if directive.validator is not None:
                    existing.validator = directive.validator

            else:
                self.validators[name] = directive

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def remove_expectation(self, name):
        """ set an expection's active to False, disabling it """
        try:
            self.set_expectation(name, active=False)
        except (KeyError,) as e:  # pragma: no cover
            logger.warning(e)
        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise

    def set_expectation(
        self,
        name,
        exp=undefined,
        validator: "Validator" = None,
        active: Optional[bool] = None,
    ):
        """ 
            note:  set_expectations is the instance-level method, add_directive works at class-level

            besides the name any of the other parameters can be unspecified
            however, come check_expectations, any `active` directive needs to 
            have a `validator` and `exp` set.

            :param name: logical name, ex: 'title'  it is used to allow successive
            adjustments to a given validation

            :param exp:  an expectation, which can pretty anything, a value, a regex to match, a callable...

            :param active: turns the expectation on and off.  left to None, it causes an error if unset


            going through the test hierarchy, each class copies expectations 
            from its mro and then qualifies them as needed.

            a test method has the final say, via `set_expectation`


        """

        try:

            if validator is not None:
                if not isinstance(validator, Validator):
                    raise InvalidConfigurationException(
                        "%s. validator %s is not an instance of Validator"
                        % (name, validator)
                    )

            return self.add_directive(
                name=name, exp=exp, validator=validator, active=active, final=True
            )

        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise

    def add_directive(
        self,
        name: str,
        exp: Any = undefined,
        validator: Validator = None,
        active: Optional[bool] = None,
        final=False,
    ):
        """ hmmmm, looks a lot like set_expectations """

        try:

            if isinstance(name, ValidationDirective):
                self._add_baseline(name)
                return

            # bit of shuffling around to fix likely config mistakes...
            if validator is None and isinstance(exp, Validator):
                validator = exp
                exp = undefined

            if isinstance(name, ValidationManager):
                for name, directive in name.validators.items():
                    # remember, you need to copy shared directives because set_expectations modifies them
                    self._add_baseline(directive.copy())
                return

            assert validator is None or isinstance(validator, Validator)

            assert not isinstance(exp, Validator)

            if active is None and exp is not undefined:
                active = True

            directive = ValidationDirective(name, validator, exp, active)
            self._add_baseline(directive)

        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if rpdb() or cpdb():
                ppp(locals())

                pdb.set_trace()
            raise

    def debug(self):
        try:
            padder = "⚙️" * 6 + "  lazy-tests configuration  " + "⚙️" * 6
            print(f"\n{padder}\n{self}  validators:")
            for key in self.validators:
                value = self.validators[key]
                print(
                    f"  {key:20.20}: active:{value.active} exp:{str(value.exp) if not isinstance(value.exp, type) else value.exp.__name__:10.10} {value.validator.__class__.__name__:30.30}"
                )

            li_revmro = list(reversed(self.testee.__class__.mro()))

            li_info = [
                f"{cls.__module__}.{cls.__name__}"
                for cls in li_revmro
                if getattr(cls, "cls_validators", None)
            ]

            print("\n\n ⚙️class-level inheritance:\n%s\n" % "\n".join(li_info))

            for cls in li_revmro:
                cls_validators = getattr(cls, "cls_validators", None)
                # print(f" {cls.__name__}:{cls_validators}")
                if cls_validators:
                    print(f"\nfrom class {cls.__name__}:")
                    if isinstance(cls_validators, list):

                        for item in cls_validators:
                            if hasattr(item, "validators"):
                                print(f"   {item.validators:60.60}")
                            else:
                                print(f"   {item}")

                    elif hasattr(cls_validators, "validators"):
                        print(f"   {cls_validators.validators:60.60}")
                    else:
                        print(f"   {cls_validators:60.60}")

            print(f"\n{padder}\n")

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if 1 or cpdb():
                pdb.set_trace()
            raise


def build_validators_for_class(cls, li_revmro):
    """build filters for class, before instance overrides"""
    try:

        classname = cls.__name__

        final = []

        seen = set()

        for cls_ in li_revmro:

            validators = getattr(cls_, "cls_validators", None)

            if validators:

                # now, what do we accept?
                if not isinstance(validators, list):
                    validators = [validators]

                validators = validators.copy()

                for validator in validators:
                    if not isinstance(
                        validator, (ValidationManager, ValidationDirective)
                    ):
                        raise InvalidConfigurationException(
                            f"unexpected validation {validator} on class.{cls_.__name__}"
                        )

                    if not validator in seen:
                        final.append(validator)
                        seen.add(validator)

                # final.extend(validators)

        return final

    # pragma: no cover pylint: disable=unused-variable
    except (Exception,) as e:
        if cpdb():
            pdb.set_trace()
        raise



import pdb
import json

#######################################################
# Dependencies
#######################################################


from .common import UnavailableLibrary

try:
    from bs4 import BeautifulSoup as bs

# pragma: no cover pylint: disable=unused-variable
except (ImportError,) as e:

    class Foo:
        pass

    # this will throw an InvalidConfigurationError on any access to bs
    # telling you to install BeautifulSoup
    bs = UnavailableLibrary(name=Foo.__module__, missing="Beautifulsoup")


from .validators import (
    ValidationManager,
    ValidationDirective,
    NamedTesteeAttributeValidator,
    FullyQualifiedNamesValidator,
    MixinExpInGot,
    Validator,
    AttrNamedDictValidator,
)

from .filters import FilterManager, RawFilter, DataMatcher

from traceback import print_exc as xp

undefined = NotImplemented


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb


from lazy_regression_tests.utils import first, fill_template, ppp, getpath

#################################################################
# the Filters
#################################################################


class CSSRemoveFilter(RawFilter, DataMatcher):
    """ this belongs in the http_validators module """

    def __init__(self, pattern, name, scalar=False, *args):
        self.selector = pattern
        self.name = name
        self.scalar = scalar

    def pre_format(self, data):
        try:
            data = bs(data)
            return data
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def filter(self, options, tmp, data, callback):
        try:

            if isinstance(data, str):
                data = self.pre_format(data)

            li = []

            for hit in data.select(self.selector):

                s_hit = str(hit)
                self.add_to_filter_hit(tmp, hit)
                li.append(hit)

            if callback:
                callback(self.name, data, li)

            for hit in li:
                hit.decompose()

            return data

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class HeadersValidator(AttrNamedDictValidator):
    """specifies source for HTTP headers """

    sourcename = "response.headers"


class ContentTypeValidator(
    FullyQualifiedNamesValidator, MixinExpInGot, HeadersValidator
):
    """ check response headers for content type.   
        not that `exp` is checked for being within the actual content type, not for equality
        exp:json matches application/json
    """

    selector = "content-type"


class StatusCodeValidator(NamedTesteeAttributeValidator):
    """ check an http status code .  you can provide 
        multiple codes in exp, ex: `exp = [200,302, "404"]`
    """

    sourcename = "response"
    selector = "status_code"

    def test(self, testee, exp, got, message, name=None):
        try:

            if message is None:
                message = fill_template(
                    "%(name)s exp:%(exp)s<>%(got)s:got",
                    locals(),
                    self,
                    {"name": name or self},
                )

            if isinstance(exp, (int, str)):
                exp = [exp]

            exp = [int(exp) for exp in exp]
            got = int(got)

            testee.assertTrue(got in exp, message)

        except (AssertionError,) as e:  # pragma: no cover
            raise

        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise


class CSSValidator(Validator):
    """ responsible for getting DOM data out, typically via Beautifulsoup
        but anything provising CSS selectors could be used.

        this class expects response to be in sources during check_expectations calls
        on that it wants a selectable attribute with select on it.

        calling code may look like this.

            response = request.get(url)
            response.selectable = BeautifulSoup(response.content)
            self.check_expectations(response=response)

    """

    sourcename = "response.selectable"

    to_text = False

    def __init__(self, selector, scalar=None, to_text=True, cargo=None):

        if scalar is None:
            scalar = selector.startswith("#")

        super(CSSValidator, self).__init__(
            selector, scalar=scalar, sourcename=self.sourcename, cargo=cargo
        )
        self.to_text = to_text

    def get_value(self, source):

        found = source.select(self.selector)
        if self.scalar:
            found = first(found)
            if found and self.to_text:
                return found.text.strip()
            return found
        else:
            if not self.to_text:
                return found
            else:
                return [elem.text.strip() for elem in found]


class TitleCSSValidator(FullyQualifiedNamesValidator, CSSValidator):
    selector = "title"
    scalar = True
    to_text = True


#######################################################
# Compose base validations/expectations
#######################################################


class HTTPValidationMixin:
    """ sets basic expectations 
        - http is expected to return a status_code, typically 200 (exp can be changed later)
        - and has a content_type, which changes depending on end points
    """

    cls_validators = [
        ValidationDirective("status_code", exp=200, validator=StatusCodeValidator()),
        ValidationDirective(
            "content_type", active=True, validator=ContentTypeValidator()
        ),
    ]


#######################################################
# HTML-type validations
#######################################################

title_validation = ValidationDirective(
    "title", active=True, validator=TitleCSSValidator()
)


class HTMLValidationMixin(HTTPValidationMixin):
    """ set `content_type` expectations
        always want to validate `title`, by default 
    """

    cls_validators = [title_validation, ValidationDirective("content_type", exp="html")]


class JSONValidationMixin(HTTPValidationMixin):
    """ sets the expected content type to JSON """

    cls_validators = [ValidationDirective("content_type", exp="json")]


class HtmlFilterManager(FilterManager):
    """Filter Manager for HTML"""

    def prep(self, tmp, data):
        try:
            if hasattr(data, "select"):
                return data
            else:
                return bs(data)
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def to_text(self, tmp, data):
        try:
            # pdb.set_trace()
            return data.prettify()
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class ResponseHTML:
    def __repr__(self):
        return "%s" % (self.__class__.__name__)

    _selectable = undefined

    @property
    def selectable(self):

        if self._selectable is undefined:
            self._selectable = bs(self.content)
        return self._selectable

    wanteds = dict(headers=["headers", "_headers"])

    def __init__(self, response):
        try:
            self._response = response
            self.content = response.content
            self.status_code = response.status_code

            for attrname, paths in self.wanteds.items():
                if isinstance(paths, str):
                    paths = [paths]

                # pdb.set_trace()
                for path in paths:
                    try:
                        value = getpath(response, path)
                        setattr(self, attrname, value)
                        break
                    # pragma: no cover pylint: disable=unused-variable
                    except (AttributeError,) as e:
                        pass
                else:
                    raise AttributeError(attrname)

            # pdb.set_trace()
            if rpdb():  # pragma: no cover
                pdb.set_trace()

        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise



# -*- coding: utf-8 -*-
"""
test json_schema  - requires `jsonschema`
"""
import pdb
import sys

import unittest


# pylint: disable=unused-import
#######################################################
# Typing
#######################################################
from typing import (
    Optional,
    # TYPE_CHECKING,
    Any,
    cast,
)

#######################################################
# pylint: enable=unused-import


verbose = "-v" in sys.argv


# pylint: disable=attribute-defined-outside-init

# pylint: disable=missing-function-docstring,missing-class-docstring  #🧨🧨🧨🧨 turn this back on later 🧨🧨🧨🧨


from lazy_regression_tests.helper_tst import get_mock_env


import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)
from traceback import print_exc as xp  # pylint: disable=unused-import


from lazy_regression_tests._baseutils import set_cpdb


from lazy_regression_tests import LazyMixin

from lazy_regression_tests import ValidationDirective


from lazy_regression_tests.json_schema_validators import JsonSchemaValidator
from lazy_regression_tests.filters import JsonFilterManager


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb


##########################################################
# a Basic JSON Schema validation
##########################################################

valid_auth_display = {
    "type": "object",
    "properties": {
        "authorizedactions": {"type": "integer"},
        "displayonly": {"type": "integer"},
    },
    "required": ["authorizedactions", "displayonly"],
}


# pylint: disable=no-member
class LazyMixinBasic(LazyMixin):

    """
    bring in basic lazytest behavior 
    not testing with assert_exp however.
    """

    data = dict(authorizedactions=3, displayonly=0)

    cls_filters = dict(json=JsonFilterManager())
    lazy_filename = LazyMixin.get_basename(__name__, __file__, __module__)

    extension = "json"

    exp_fail = False

    def test_it(self):
        """ call the validations

        - if exp_fail is set, 
          -  make sure you get an AssertionError 
          -  and that `exp_fail` appears in the assertion error message

        - otherwise behave as usual, i.e. fail on an assertion error

        """
        try:
            data = self.data.copy()

            if not self.exp_fail:
                self.check_expectations(data=data)
            else:
                try:
                    self.check_expectations(data=data)
                    self.fail(f"expected {self.exp_fail}")
                # pragma: no cover pylint: disable=unused-variable
                except (AssertionError,) as e:
                    self.assertTrue(
                        str(self.exp_fail) in str(e), f"expected {self.exp_fail} in {e}"
                    )

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class Test_AuthDisplay(LazyMixinBasic, unittest.TestCase):

    data = dict(authorizedactions=3, displayonly=0)

    cls_validators = [
        ValidationDirective(
            "valid1",
            exp=valid_auth_display,
            validator=JsonSchemaValidator(),
            active=True,
        )
    ]


class Test_AuthDisplayBad(Test_AuthDisplay):

    displayonly = "XXX"
    data = dict(authorizedactions=3, displayonly=displayonly)
    exp_fail = displayonly


valid_carrots = {
    "type": "object",
    "properties": {"carrots": {"type": "integer"}},
    "required": ["carrots"],
}


class Test_MultiChecks(Test_AuthDisplay, unittest.TestCase):

    cls_validators = [
        ValidationDirective(
            "valid_self",
            exp=valid_carrots,
            validator=JsonSchemaValidator(sourcename=None, selector="pantry.veggies"),
            active=True,
        )
    ]

    def setUp(self):
        self.pantry = dict(veggies=dict(carrots=5))


class Test_MultiCheckFail(Test_MultiChecks):

    displayonly = "XXX"
    data = dict(authorizedactions=3, displayonly=displayonly)
    exp_fail = displayonly


if __name__ == "__main__":

    cpdb = set_cpdb()

    sys.exit(unittest.main())



""" core LazyMixin functionality 

be very cautious of referencing self.<attribute> when there is no 'lazy' prefix to it

for this reason auxiliary classes have been moved to core_assist.py

"""

import pdb
import os
import sys

from pathlib import Path

undefined = NotImplemented

verbose = "-v" in sys.argv

#######################################################
# Typing
#######################################################
from typing import (
    Optional,
    # TYPE_CHECKING,
    Any,
    Union,
)


#######################################################


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb


try:

    from timeout_decorator import timeout

    import timeout_decorator

    CustomTimeoutError = timeout_decorator.timeout_decorator.TimeoutError

    TIMEOUT_MAXTIME_TO_ALLOW = 5
except (ImportError,) as e:  # pragma: no cover
    # pdb.set_trace()
    timeout = None
    TIMEOUT_MAXTIME_TO_ALLOW = 0

    class CustomTimeoutError(Exception):
        """we'll never see this """


# pylint: disable=unused-import
from lazy_regression_tests._baseutils import ppp, Dummy
from traceback import print_exc as xp

# pylint: enable=unused-import


# pylint: disable=unused-wildcard-import,wildcard-import
from .common import *

# pylint: enable=unused-wildcard-import,wildcard-import


from .validators import ValidationManager


# aliasing the JSON response filter management to DictFilterManager as there is
# very little that is HTTP specific

from lazy_regression_tests.utils import Subber, fill_template


from .core_assist import LazyChecker, LazyTemp, MediatedEnvironDict, _Control, _LazyMeta


class LazyMixin(metaclass=_LazyMeta):
    """ Intended as a Mixin to unittest classes it provides a number of services automatically

        - validations via cls_validators and set_expectation
        - regression testing against last seen baseline via `self.assert_exp(data, extension)`
        - filtering to avoid false regression alerts from variable data (timestamps, csrf protection tokens)
           - cls_filters = dict(<extension>=FilterDirectives...)
           - self.set_filter(FilterDirective)

        - both validations and filters are intended to be defined in mixin ancestor classes and mixed and 
          matched as desired

        class MyTest(LazyMixin, unittest.TestCase):
            cls_filters = dict(html=filter_csrftokens)
            cls_validators = [validate_http, validate_html]

            
    """

    cls_filters = {}
    cls_validators = []
    add_lazy_dirname = []

    lazy_dirname_extras = []

    lazytemp = None
    lazy_basename_extras = ""

    # this normally resolves to os.environ, but can be preset for testing
    lazy_environ = MediatedEnvironDict()

    T_FILENAME = "%(filename)s %(classname)s %(_testMethodName)s %(lazy_basename_extras)s %(suffix)s %(lazy_extension)s"

    ENVIRONMENT_VARNAME_ROOT = "lzrt_"

    @classmethod
    def get_basename(cls, name_, file_, module_) -> str:
        """ called from mytest.py this returns mytest and is the base for file naming
        its use in client code is a bit anomalous as the enclosing class
        hasn't been created yet.
        """
        cls.lazy_pa = pa = Path(file_)
        return pa.stem

    def lazy_build_filters(self):
        try:
            res = {}
            for extension, filter_ in self.__class__.cls_filters.items():
                # if cpdb(): pdb.set_trace()
                filter_copy = filter_.copy()
                assert filter_copy, "filter_copy empty"
                res[extension] = filter_copy

            return res

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def _lazy_get_t_dirname(self, exp_got, subber):
        """ get the template for the directory names where to save files """
        try:

            env_name = dict(
                exp="template_dirname_exp",
                got="template_dirname_got",
                report="template_dirname_report",
            )[exp_got]
            dirname = subber.get("dirname") or self._lazy_control.env.get(env_name)

            if dirname is None:
                if exp_got in ("exp", "got"):
                    raise InvalidConfigurationException(
                        "could not get output directory for %s in %s"
                        % (exp_got, self._lazy_control.env)
                    )
                return None

            dirname2 = os.path.join(
                dirname, self.lazy_filename, subber.get("classname")
            )

            return dirname2

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def _handle_dirname_extras(self, _litd):
        """ allows injection of extra instance-set attributes into the directory hierarchies
        The intent of `lazy_dirname_extras` is to partition tests by other attributes, like say 
        a site name or a test database name.
        given `lazy_dirname_extras = "site"` and `self.site = 'example.com'`
        .
        └── <myclass>
            ├── example.com
            │   └── <get_basename>.<myclass>.<_testMethodName>.txt

        """

        try:

            dirname_extras = getattr(self, "lazy_dirname_extras", "")
            if not dirname_extras:
                return _litd

            if isinstance(dirname_extras, list):
                dirname_extras = " ".join(dirname_extras)
            if dirname_extras:
                # expand something like "foo, bar" into [..."%(foo)s", "%(bar)s"...]
                li_replace = [
                    "%%(%s)s" % (attrname) for attrname in dirname_extras.split()
                ]

                _litd.extend(li_replace)

            return _litd

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def _get_fnp_save(
        self,
        exp_got: Union["got", "exp"],
        options: "LazyChecker",
        suffix: Optional[str],
    ):
        """ get the save path """

        try:

            subber = Subber(
                options,
                {
                    "filename": self.lazy_filename,
                    "suffix": suffix,
                    "classname": self.__class__.__name__,
                    "exp_got": exp_got,
                },
                # the lower priority the TestCase instance the less probability
                # of name clashes
                self,
            )

            # calculating the directory path
            t_dirname = self._lazy_get_t_dirname(exp_got, subber)

            if t_dirname is None:
                return None

            _litd = t_dirname.split(os.path.sep)

            _litd = self._handle_dirname_extras(_litd)

            _lid = ["/"] + [fill_template(t_, subber) for t_ in _litd]

            dirname = os.path.join(*_lid)

            # calculating the filename
            t_basename = self.T_FILENAME
            _litb = t_basename.split()
            _lib = [fill_template(t_, subber) for t_ in _litb]
            basename = ".".join([i_ for i_ in _lib if i_])

            basename = basename.replace(" ", "_")
            basename = basename.replace("/", "_")

            if not os.path.isdir(dirname):
                os.makedirs(dirname)

            return os.path.join(dirname, basename)

        # pragma: no cover pylint: disable=unused-variable, broad-except
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    #######################################################
    # validator related
    #######################################################

    _filters = _validationmgr = undefined

    @property
    def validationmgr(self):
        """ create the validationmgr on demand """
        if self._validationmgr is undefined:
            self._validationmgr = ValidationManager(self, *self.cls_validators)

        return self._validationmgr

    @property
    def filters(self):
        """ build filters at the instance level """
        if self._filters is undefined:
            self._filters = self.lazy_build_filters()

        return self._filters

    def check_expectations(
        self, lazy_skip=None, lazy_skip_except=None, lazy_sourced_only=True, **sources
    ):
        """ validate active validation directives """
        try:
            self.validationmgr.check_expectations(
                self,
                lazy_skip=lazy_skip,
                lazy_skip_except=lazy_skip_except,
                lazy_sourced_only=lazy_sourced_only,
                **sources,
            )
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def set_expectation(self, *args, **kwargs):
        """ add/modify a validation """

        validationmgr = self.validationmgr

        name = args[0]
        if breakpoints("set_expectation", {"name": name}):  # pragma: no cover
            pdb.set_trace()

            # put this in your breakpoints.json

        validationmgr.set_expectation(*args, **kwargs)

    #######################################################
    # diff-related
    #######################################################

    def assert_exp(self, got: Any, extension: str, suffix: str = "") -> LazyTemp:
        """
        regression test that `got` is the same as what was last encountered and stored
        in `exp` file.
        """

        try:

            if not isinstance(extension, str):
                raise InvalidConfigurationException(
                    "%s.extension has to be a string (extension=%s) and one of the existing filters"
                    % (self, extension)
                )

            try:
                filter_ = self.filters[extension]
            # pragma: no cover pylint: disable=unused-variable
            except (KeyError,) as e:
                raise InvalidConfigurationException(
                    f"{self}. unknown extension={extension}. known extensions in filters:{self.filters.keys()}"
                )

            rawfiltermgr, textfiltermgr = filter_.get_raw_text_filters()

            checker = LazyChecker(
                extension=extension,
                rawfiltermgr=rawfiltermgr,
                textfiltermgr=textfiltermgr,
            )

            if hasattr(filter_, "to_text"):
                checker.to_text = filter_.to_text
            if hasattr(filter_, "prep"):
                checker.prep = filter_.prep

            return self._lazycheck(got, checker, suffix)

        # pragma: no cover pylint: disable=unused-variable
        except (AssertionError,) as e:  # pragma: no cover
            raise
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def _get_fnp_raw(self, options, suffix, control, tmp, got):
        """ 
        where do we save the raw received data,
        after formatting, but before filtering ? 
        """

        try:
            if not control.save_raw:
                return

            fnp_raw = self._get_fnp_save("report", options, suffix)

            return fnp_raw

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def _lazycheck(self, got: Any, options: LazyChecker, suffix: str = "") -> LazyTemp:

        try:
            env = self.lazy_environ
            if not self.lazy_environ.acquired:
                env.clear()
                env.acquire(self.ENVIRONMENT_VARNAME_ROOT)

            self._lazy_control = control = _Control(self, env, options)

            # only create the lazy temp the first time.
            tmp = self.lazytemp = self.lazytemp or LazyTemp(control, env, self)

            # the environment requests that no diffing or writing take place
            # typically indicated by setting environment variable `lzrt_directive=skip`
            if control.skip():
                return tmp

            # calculate the paths of the exp/got files
            tmp.fnp_got = fnp_got = self._get_fnp_save("got", options, suffix)
            tmp.fnp_exp = fnp_exp = self._get_fnp_save("exp", options, suffix)

            fnp_raw = self._get_fnp_raw(options, suffix, control, tmp, got)

            # linefeeds have a tendency to creep in sometimes
            formatted_got = options.format(tmp, got, fnp_raw).rstrip()

            # at this point, we want to write the received, formatted, data regardless
            with open(fnp_got, "w") as fo:
                fo.write(formatted_got)

            # the newly received data is to be taken as our expectations
            # typically indicated by setting environment variable `lzrt_directive=baseline`
            if control.baseline():
                with open(fnp_exp, "w") as fo:
                    fo.write(formatted_got)
                return tmp

            try:
                with open(fnp_exp) as fi:
                    # linefeeds have a tendency to creep in sometimes
                    exp = fi.read().rstrip()

            # pragma: no cover pylint: disable=unused-variable
            except (IOError,) as e:
                # we just want to write the received data as our expectation
                tmp.execinfo.ioerror_exp = fnp_exp
                tmp.message = message = "no check because IOError on %s" % (fnp_exp)
                with open(fnp_exp, "w") as fo:
                    fo.write(formatted_got)
                return tmp

            # the environment requests only equality is checked, without trying to show details
            # typically indicated by setting environment variable `lzrt_directive=nodiff`
            # this may be desired if the differences could cause timeouts with `assertEqual`
            if control.nodiff():
                tmp.message = message = "exp and got are not equal but diffing disabled"
                if exp != formatted_got():
                    raise self.fail(message)

            # pragma: no cover pylint: disable=unused-variable
            try:
                # supports a timeout mechanism, if the module is available
                self.assertEqualTimed(exp, formatted_got)
            except (CustomTimeoutError,) as e:  # pragma: no cover
                tmp.message = message = (
                    "exp and got are not equal but comparison timed out after %s seconds"
                    % (TIMEOUT_MAXTIME_TO_ALLOW)
                )
                self.fail(message)
            except (AssertionError,) as e:  # pragma: no cover
                raise
            except (Exception,) as e:
                if cpdb():
                    pdb.set_trace()
                raise

            return self.lazytemp

        except (AssertionError,) as e:  # pragma: no cover
            raise
        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise

    #######################################################
    # Note the conditional method definition and the fallback to
    # basic assertEqual
    #######################################################
    if timeout:

        @timeout(TIMEOUT_MAXTIME_TO_ALLOW)
        def assertEqualTimed(self, exp, got, message=None):
            """ comparisons will automatically times out after %s seconds""" % (
                TIMEOUT_MAXTIME_TO_ALLOW
            )
            try:
                self.assertEqual(exp, got, message)
            except (AssertionError,) as e:
                raise self.format_assertion_message(e)

    else:
        #
        def assertEqualTimed(self, exp, got, message=None):
            """ fallback if timeout package is not available """
            try:
                self.assertEqual(exp, got, message)
            except (AssertionError,) as e:
                raise self.format_assertion_message(e)

    def format_assertion_message(self, ori):

        message = f"""

❌❌❌Regressions found:

Expected contents are in file:
  {self.lazytemp.fnp_exp}

Received contents are in file:
  {self.lazytemp.fnp_got}

Original exception:
 {ori}
❌❌❌
"""
        exc = ori.__class__(message)
        return exc



# -*- coding: utf-8 -*-
"""
test core behavior - does not need any third party libraries
"""

import sys
import os
import re

import unittest

try:
    import unittest.mock as mock
except (ImportError,) as ei:
    import mock  # python 2?


# pylint: disable=unused-import
#######################################################
# Typing
#######################################################
from typing import (
    Optional,
    # TYPE_CHECKING,
    Any,
    cast,
)

#######################################################
# pylint: enable=unused-import


verbose = "-v" in sys.argv


# pylint: disable=attribute-defined-outside-init

# pylint: disable=missing-function-docstring,missing-class-docstring  #🧨🧨🧨🧨 turn this back on later 🧨🧨🧨🧨

import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)
from traceback import print_exc as xp  # pylint: disable=unused-import


from lazy_regression_tests._baseutils import (
    set_cpdb,
    set_rpdb,
    ppp,
    debugObject,
    fill_template,
    Subber,
    RescueDict,
    Dummy,
    set_breakpoints3,
)

from lazy_regression_tests.utils import InvalidConfigurationException


from lazy_regression_tests import (
    DictValidator,
    ValidationDirective,
    DirectValidator,
    AutoExp,
)


from lazy_regression_tests.filters import (
    RegexRemoveSaver,
    DictFilter,
    FilterDirective,
    FilterManager,
    JsonFilterManager,
)

from lazy_regression_tests.core import OPT_DIRECTIVE_BASELINE


rescuedict = RescueDict()

import pdb


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb


from lazy_regression_tests import LazyMixin

##########################################################
# tests
##########################################################


from lazy_regression_tests.helper_tst import get_mock_env


di_mock_env = get_mock_env()

lzrt_template_dirname_got = di_mock_env["lzrt_template_dirname_got"]
lzrt_template_dirname_exp = di_mock_env["lzrt_template_dirname_exp"]


di_mock_env_baseline = di_mock_env.copy()
di_mock_env_baseline.update(lzrt_directive=OPT_DIRECTIVE_BASELINE)


di_mock_env_no_extras = dict(
    [(k, v.replace("/%(lazy_dirname_extras)s", "")) for k, v in di_mock_env.items()]
)

module_ = "builtins"
module_ = module_ if module_ in sys.modules else "__builtin__"
funcpath_open = "%s.open" % module_


def debug_env(self):
    ppp(self.lazy_environ, "lazyenv")


# pylint: disable=no-member
class LazyMixinBasic(LazyMixin):

    """
    basic support functionality for the rest of the tests
    similar in fact to how LazyMixin is normally used.

    """

    cls_filters = dict(txt=FilterManager())

    debug_env = debug_env

    lazy_filename = LazyMixin.get_basename(__name__, __file__, __module__)

    tmp_formatted_data = None

    extension = "txt"

    data = """
somedata
var1
var2
"""

    def setUp(self):
        """not sure why needed, but need to reset the environment
            to make sure mocked values are taken into account
           This is probably due to some too aggressive optimization,
           i.e. we-dont-expect-env-variables to change during a test run
        """

        self.lazy_environ.acquired = False

    def seed(self, exp, extension=None, suffix=""):
        try:

            tmp = getattr(self, "lazytemp", None)

            exception = None
            try:

                extension = extension or self.extension
                self.assert_exp(exp, extension=extension, suffix=suffix)
            # pragma: no cover pylint: disable=unused-variable
            except (AssertionError,) as e:
                pass

            tmp = tmp or getattr(self, "lazytemp", None)
            self.lazytemp = None
            return tmp

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def check_naming_convention(self):
        """        
        (under lzrt_template_dirname_exp/lzrt_template_dirname_got directory roots)        
        👇 exp and got files are written to $lzrt_template_dirname_exp, $lzrt_template_dirname_got 
           environment variables
        <root dir exp/got>
        |  
        └──test_core/        👈 module 
            └──TestBasic/    👈 class
               ├── test_core.TestNamingConventions.test_001_base_naming_convention.txt
               └── test_core.TestNamingConventions.test_002_suffix.mysuffix.txt
                             👆                    👆                       👆 
                              class                method                   extension


        It is also possible to inject one or more extra parts into the directory,
        above, the classname this is done via the `lazy_dirname_extras` 
        variable which results in an attribute lookup on the TestCase instance
        that is inserted into the path

        class MyTestClass:
            lazy_dirname_extras = ["site"]

            def test_site(self):
                self.site = "example.com"
    
        TestNamingConventions/
        ├── example.com
            └── test_core.TestNamingConventions.test_003_dirname_extras_list.txt    

        """
        try:
            t_msg = ":%s: not found in :%s:"

            classname = self.__class__.__name__

            for fnp, dir_root in zip(
                [self.lazytemp.fnp_got, self.lazytemp.fnp_exp],
                [lzrt_template_dirname_got, lzrt_template_dirname_exp],
            ):
                self.assertTrue(fnp.endswith(".%s" % (self.extension)))

                self.assertTrue(fnp.startswith(dir_root))

                # check that the module name is in the directories
                dirname, fname = os.path.split(fnp)
                dirs_hierarchy = dirname.split(os.path.sep)
                msg = f"filename:{self.lazy_filename}: not found in hierarchy: {dirs_hierarchy}"
                self.assertTrue(self.lazy_filename in dirs_hierarchy, msg)

                if getattr(self, "lazy_dirname_extras", None):
                    li = self._handle_dirname_extras([])

                    li2 = [classname] + [
                        fill_template(tmpl, self, rescuedict) for tmpl in li
                    ]

                    tail = os.path.join(*li2)

                    self.assertTrue(
                        dirname.endswith(tail),
                        "dir %s does end with %s extras/class" % (dirname, tail),
                    )

                else:

                    self.assertTrue(
                        dirname.endswith(classname),
                        "dir %s does end with %s class" % (dirname, classname),
                    )

                # class name and method are both in the filename
                for part in [classname, self._testMethodName]:
                    msg = t_msg % (part, fname)
                    self.assertTrue(part in fname, msg)

                for part in [classname]:
                    msg = t_msg % (part, dirname)
                    self.assertTrue(part in dirname, msg)

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


# pylint: enable=no-member


def get_subber_for_dirname(self, di_env: Optional[dict] = None):
    try:

        di_env = di_env or {}

        di_sub = dict(suffix="")
        subber = Subber(
            self,
            di_sub,
            {
                "filename": self.lazy_filename,
                "suffix": "",
                "classname": self.__class__.__name__,
                "exp_got": "exp",
                "extension": "txt",
            },
        )

        self.control = Dummy(env=di_env.copy())

        return subber

    # pragma: no cover pylint: disable=unused-variable
    except (Exception,) as e:
        if cpdb():
            pdb.set_trace()
        raise


class TestBasic(LazyMixinBasic, unittest.TestCase):
    @mock.patch.dict(os.environ, di_mock_env)
    def test_001_equal_string(self):
        """sanity check - are we ok if exp == got"""

        try:
            self.seed(self.data, self.extension)
            self.assert_exp(self.data, self.extension)

        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise


class TestNamingConventions(LazyMixinBasic, unittest.TestCase):
    @mock.patch.dict(os.environ, di_mock_env)
    def test_001_base_naming_convention(self):

        try:

            self.seed(self.data, self.extension)
            self.assert_exp(self.data, self.extension)
            self.check_naming_convention()

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_mock_env)
    def test_002_suffix(self):
        """
        TestBasic/
        └── test_lazymixin.TestBasic.test_003_suffix.mysuffix.txt
                    ends with .suffix.extension      👆 
        """

        suffix = "mysuffix"

        self.seed(self.data, self.extension, suffix=suffix)
        self.assert_exp(self.data, self.extension, suffix=suffix)

        self.check_naming_convention()

        t_msg = ":%s: does not end with :%s:"

        for fnp in [self.lazytemp.fnp_got, self.lazytemp.fnp_exp]:
            _, fname = os.path.split(fnp)
            should_end = ".%s.%s" % (suffix, self.extension)

            msg = t_msg % (fname, should_end)

            self.assertTrue(fname.endswith(should_end), msg)

    @mock.patch.dict(os.environ, di_mock_env)
    def test_003_dirname_extras_list(self):
        """
        
        The intent of `lazy_dirname_extras` is to partition tests by other attributes, like say 
        a site name or a test database name.

        .
        └── TestNamingConventions
            ├── example.com
            │   └── test_lazymixin.TestNamingConventions.test_003_dirname_extras_list.txt

        """

        self.lazy_dirname_extras = ["site"]

        self.site = "example.com"

        self.seed(self.data, self.extension)
        self.assert_exp(self.data, self.extension)

        self.check_naming_convention()

    @mock.patch.dict(os.environ, di_mock_env)
    def test_004_dirname_multiple_string(self):
        """
        .
        └── TestNamingConventions
            ├── example.com
            │   ├── testB
            │   │   └── test_lazymixin.TestNamingConventions.test_004_dirname_multiple_string.txt
        """

        self.lazy_dirname_extras = "site abtest"

        self.site = "example.com"
        self.abtest = "testB"

        self.seed(self.data, self.extension)
        self.assert_exp(self.data, self.extension)

        self.check_naming_convention()


lorem = """
<html><head><title></title></head><body>
<p>Lorem ipsum dolor sit amet, consectetur 
adipisicing elit. Consequatur Quidem, sint incidunt?</p></body></html>
"""


class Test_Error_Handling(LazyMixinBasic, unittest.TestCase):
    """
    if `self.lazytemp.fnp_exp` is not found, that means this is a first time run
    and `got` is written to both `self.lazytemp.fnp_exp` and `self.lazytemp.fnp_got`
    and assert_exp is automatically considered to be successful

    this is also the behavior with $lzrt_directive == "baseline" environment variable
    """

    @mock.patch.dict(os.environ, di_mock_env)
    def test_001_exp_does_not_exist(self):

        try:

            # first time run, so exp does not exist
            self.assert_exp(self.data, self.extension)

            for fnp in [self.lazytemp.fnp_got, self.lazytemp.fnp_exp]:
                with open(fnp) as fi:
                    data = fi.read()
                self.assertEqual(self.data.strip(), data.strip())

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    #

    @mock.patch.dict(os.environ, di_mock_env_baseline)
    def test_002_baseline(self):
        """
        the environment has been mocked to tell lazy to consider the data
        as a baseline:

        - `got` is written to both `self.lazytemp.fnp_exp` and `self.lazytemp.fnp_got`
        - assert_exp is automatically considered to be successful

        """

        try:

            # we are changing the data to go into the seed, i.e. the `exp` side of things
            changed_data = self.data.replace("var2", "var3")

            self.assertFalse(changed_data.strip() == self.data.strip())

            lazytemp = self.seed(changed_data, self.extension)

            for fnp in [lazytemp.fnp_got, lazytemp.fnp_exp]:
                with open(fnp) as fi:
                    data = fi.read()
                self.assertEqual(changed_data.strip(), data.strip())

            # now we call assert_exp, but the data will be taken as baseline
            self.assert_exp(self.data, self.extension)

            for fnp in [self.lazytemp.fnp_got, self.lazytemp.fnp_exp]:
                with open(fnp) as fi:
                    data = fi.read()
                self.assertEqual(self.data.strip(), data.strip())

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_mock_env)
    def test_003_changed_data(self):
        """
        Here the data is changed, but the exp is already seeded and
        lzrt_directive is unset, i.e. it is not 'baseline'
        An assertionError should be thrown
        """

        try:

            # we are changing the data to go into the seed, i.e. the `exp` side of things
            changed_data = self.data.replace("var2", "var3")

            self.assertFalse(changed_data.strip() == self.data.strip())

            lazytemp = self.seed(changed_data, self.extension)

            for fnp in [lazytemp.fnp_got, lazytemp.fnp_exp]:
                with open(fnp) as fi:
                    data = fi.read()
                self.assertEqual(changed_data.strip(), data.strip())

            # expect an error now
            try:
                self.assert_exp(self.data, self.extension)
                self.fail("should have thrown an AssertionError")
            # pragma: no cover pylint: disable=unused-variable
            except (AssertionError,) as e:
                pass

            # it would be nice to re-mock os.environ and with
            # $lzrt_directive=baseline...  that should then pass

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


filtername23 = "var2and3"

filter_23 = FilterDirective(filtername23, RegexRemoveSaver("var2|var3", "var2and3"))

filtername1 = "var1"

filter_1 = FilterDirective(filtername1, RegexRemoveSaver("var1", "var1"))


class Test_Text_Filtering(LazyMixinBasic, unittest.TestCase):
    """
    filters are additive through a class's mro
    """

    filtername = filtername23

    cls_filters = dict(txt=filter_23)

    @mock.patch.dict(os.environ, di_mock_env)
    def test_001_filter(self):
        try:

            changed_data = self.data.replace("var2", "var3")

            li_change = ["var2", "var3"]

            fnp_exp = self.seed(changed_data, self.extension).fnp_exp
            self.lazytemp = None

            # test that lines matching the filter are removed from the file
            with open(fnp_exp) as fi:
                data = fi.read()

            for text in li_change:
                self.assertFalse(
                    text in data,
                    "%s should have been stripped out of %s" % (text, data),
                )

            # but the filtered data is still accessible through the filtered dictionary
            # tmp = self.lazytemp
            # # this is the value originally passed into seed
            # self.assertEqual(tmp.filtered[self.filtername], ["var3"])

            # # tmp is reset on each new instance but we are still on the same instance after seed
            # # by default the filter results are put in a list
            # tmp.filtered[self.filtername] = []

            tmp = self.assert_exp(self.data, self.extension)

            with open(tmp.fnp_got) as fi:
                data = fi.read()

            for text in li_change:
                self.assertFalse(
                    text in data,
                    "%s should have been stripped out of %s" % (text, data),
                )

            # this the unchanged value in self.data
            self.assertEqual(["var2"], tmp.filtered[self.filtername])

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_mock_env)
    def test_002_filter_scalar(self):
        """
        instead of the standard, list-appending filtering we want to have a scalar result
        to do this we can modify the filter
        """

        try:

            scalar_filter = RegexRemoveSaver("var2|var3", "var2and3", scalar=True)

            # we can modify a filter if we know its name
            self.filters[self.extension].set_filter(
                self.filtername, filter_=scalar_filter
            )

            # did we change the filter?
            newfilter = self.filters[self.extension].filters[self.filtername].filter_

            self.assertTrue(scalar_filter is newfilter)

            # and it should be a scalar
            self.assertEqual(True, newfilter.scalar)

            changed_data = self.data.replace("var2", "var3")

            li_change = ["var2", "var3"]

            tmp = self.seed(changed_data, self.extension)
            fnp_exp = tmp.fnp_exp

            # test that lines matching the filter are removed from the file
            with open(fnp_exp) as fi:
                data = fi.read()

            for text in li_change:
                self.assertFalse(
                    text in data,
                    "%s should have been stripped out of %s" % (text, data),
                )

            # but the filtered data is still accessible through the filtered dictionary

            # this is the value originally passed into seed, as a scalar
            self.assertEqual(tmp.filtered[self.filtername], "var3")

            tmp = self.assert_exp(self.data, self.extension)

            with open(tmp.fnp_got) as fi:
                data = fi.read()

            for text in li_change:
                self.assertFalse(
                    text in data,
                    "%s should have been stripped out of %s" % (text, data),
                )

            # this the unchanged value in self.data
            self.assertEqual(tmp.filtered[self.filtername], "var2")

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


#################################################################
# Test MRO-based additive filtering
# filters go from most-generic to most-specific, with named filter
# overriding precedings settings.
#################################################################


class AnyName:
    cls_filters = dict(txt=filter_23)


class AnotherName:
    cls_filters = dict(txt=filter_1)


class Test_Additive_Filtering(AnyName, AnotherName, LazyMixinBasic, unittest.TestCase):
    def change_data(self, data):
        li_change = ["var1", "var2", "var3"]

        changed_data = data
        for value in li_change:
            changed_data = changed_data.replace(value, "%sxxx" % (value))

        return changed_data

    @mock.patch.dict(os.environ, di_mock_env)
    def test_001_everything_filteredout(self):
        try:

            changed_data = self.change_data(self.data)
            self.seed(changed_data, self.extension)

            self.assert_exp(self.data, self.extension)

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_mock_env)
    def test_002_deactivated_filter_fail(self):
        try:

            changed_data = self.change_data(self.data)

            self.filters[self.extension].set_filter(filtername1, active=False)
            self.seed(changed_data, self.extension)

            try:
                self.assert_exp(self.data, self.extension)
                self.fail("should have failed on var1 difference")
            # pragma: no cover pylint: disable=unused-variable
            except (AssertionError,) as e:
                self.assertTrue("var1" in str(e))

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_mock_env)
    def test_003_bad_activation(self):
        """
        you cant activate a filter that doesn't exist
        the error happens when assert_exp is called and filtering is attempted
        this supports the scenario where:

        # we know we need to filter out changing data, but we let the subclasses
        # decide how to do it
        cls_filters = dict(txt, FilterDirective("changing",active=True))

        #
        cls_filters = dict(txt, FilterDirective("changing",filter_=<some filter>))


        """
        try:

            # dont care so much if you try to deactivate a filter that doesnt exist
            self.filters[self.extension].set_filter("xyz", active=True)

            try:

                self.assert_exp(self.data, self.extension)
                self.fail("should have failed on due to xyz filter not existing")

            # pragma: no cover pylint: disable=unused-variable
            except (InvalidConfigurationException,) as e:
                self.assertTrue("xyz" in str(e))

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_mock_env)
    def test_004_tolerate_bad_deactivation(self):
        """
        you can deactivate a filter that has not been set
        """
        try:

            # dont care so much if you try to deactivate a filter that doesnt exist
            self.filters[self.extension].set_filter("xxx", active=False)
            self.assert_exp(self.data, self.extension)

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class BaseForJson:

    cls_filters = dict(json=JsonFilterManager())

    j_data = dict(
        var1=1,
        var2=dict(var21=21, var22=22, var23=dict(var231=231)),
        var3=3,
        var4=4,
        someval="xyz",
    )


class Test_JSON(BaseForJson, LazyMixinBasic, unittest.TestCase):

    extension = "json"

    def setUp(self):
        self.j_data = self.j_data.copy()
        self.lazy_environ.acquired = False

    @mock.patch.dict(os.environ, di_mock_env)
    def test_001_same(self):
        """sanity check"""
        try:
            self.seed(self.j_data, self.extension)
            self.assert_exp(self.j_data, self.extension)
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    changes = dict(var1="one")

    def change_data(self, changes=None):

        changes = changes or self.changes.copy()

        j_data = self.j_data.copy()
        j_data.update(**changes)
        return j_data

    @mock.patch.dict(os.environ, di_mock_env)
    def test_002_changes(self):
        """sanity check"""
        try:
            changed_data = self.change_data()
            self.seed(changed_data, self.extension)
            try:
                self.assert_exp(self.j_data, self.extension)
                self.fail("should have changed on %s" % (self.changes))

            # pragma: no cover pylint: disable=unused-variable
            except (AssertionError,) as e:
                # pdb.set_trace()
                self.assertTrue("var1" in str(e))

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class Test_JSON_DictFilter(Test_JSON):

    """
    a DictFilter is a way modify incoming data and takes a dictionary of its own
    as a specifier

    #anything found in a matching location with None is just taken out.
    var1=None

    #and a dictionary value recurses
    var2=dict(var22=None)

    """

    def setUp(self):
        self.j_data = self.j_data.copy()
        self.lazy_environ.acquired = False

    di_varying = dict(var1=None, var2=dict(var22=None))

    cls_filters = dict(
        json=FilterDirective("changing", DictFilter(di_varying, "changing"))
    )

    @mock.patch.dict(os.environ, di_mock_env)
    def test_002_changes(self):
        """sanity check"""
        try:

            changed_data = self.change_data()
            changed_data["var2"]["var22"] = "changed"

            fnp_exp = self.seed(changed_data, self.extension).fnp_exp
            with open(fnp_exp) as fi:
                data = fi.read()

            # should still be there...
            s_unwanted = {"var1", "var22"}

            s_wanted = set(self.j_data.keys()) - s_unwanted

            for key in s_wanted:
                self.assertTrue(key in data)

            for key in s_unwanted:
                self.assertFalse(key in data)

            self.assert_exp(self.j_data, self.extension)

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_mock_env)
    def test_003_text_stillworks(self):
        """
        `assert_exp` is built around extensions, on purpose.  
        You can run tests on several data formats.
        """

        try:

            self.seed(self.j_data, self.extension)

            self.assert_exp(self.j_data, self.extension)
            self.check_naming_convention()

            fnp_exp = self.seed(self.data, "txt").fnp_exp
            self.assertTrue(fnp_exp.endswith(".txt"))
            self.assert_exp(self.data, "txt")

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class CustomDictValidator(DictValidator):
    """  let's write our own

         for keys prefixed with `<selector>`, 'var' in this example 
         check that value ends with what comes after the prefix
         numeric if applicable

         var2 : 2 ✅
         var3 : 5 ❌
         varfoo : "foo" ✅
    """

    def check(self, name: str, testee: "LazyMixin", exp: Any, sources: dict):
        """

        """

        try:
            source_ = self.get_source(testee, **sources)
            got = self.get_value(source_)

            for key, value in got:

                expvalue = key.split(self.selector)[1]
                try:
                    expvalue = int(expvalue)
                # pragma: no cover pylint: disable=unused-variable
                except (ValueError,) as e:
                    pass

                testee.assertEqual(expvalue, value)

        except (AssertionError,) as e:  # pragma: no cover
            raise
        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise

    def get_value(self, source, res=None):
        """ we are getting a dict and we recurse into 
            because this is what this validator wants to do
            we could have used any number of data selection
            strategies
        """
        try:
            if res is None:
                res = []
            for key, value in source.items():
                if key.startswith(self.selector):
                    if isinstance(value, dict):
                        self.get_value(value, res)
                    else:
                        res.append((key, value))
            return res
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class Test_JSON_Validation(BaseForJson, LazyMixinBasic, unittest.TestCase):

    cls_validators = [
        ValidationDirective("var3", DictValidator("var3"), exp=3),
        ValidationDirective("nested", DictValidator("var2.var23.var231"), exp=231),
    ]

    def test_001_expects(self):
        try:
            self.check_expectations(data=self.j_data)
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_002_fail231(self):
        try:
            self.set_expectation("nested", exp=232)
            try:
                self.check_expectations(data=self.j_data)
                self.fail("should have errored 231<>232")
            except (AssertionError,) as e:
                self.assertTrue("232" in str(e) and "231" in str(e))

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_003_add_expected_fail(self):
        try:
            self.set_expectation(
                ValidationDirective("var22", DictValidator("var22"), exp=23)
            )
            try:
                self.check_expectations(data=self.j_data)
                self.fail("should have errored 22<>23")
            except (AssertionError,) as e:
                self.assertTrue("22" in str(e) and "23" in str(e))

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_004_custom(self):
        try:
            self.set_expectation(
                "custom", validator=CustomDictValidator("var"), exp="N/A", active=True
            )
            self.check_expectations(data=self.j_data)

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_005_deactivate(self):
        try:
            data = self.j_data.copy()
            data["var3"] = "not3"

            # since we know we will fail the var3=3 test, we turn it off

            self.set_expectation("var3", active=False)

            self.check_expectations(data=data)

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_006_change_expectations(self):
        try:
            data = self.j_data.copy()
            data["var3"] = "not3"

            # instead of deactivateing the var3 check, we change its expectations

            self.set_expectation("var3", exp="not3")

            self.check_expectations(data=data)

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_007_cant_activate_missing_validation(self):
        try:

            self.set_expectation("no_such_validation", active=True)

            try:
                self.check_expectations(data=self.j_data)
            # pragma: no cover pylint: disable=unused-variable
            except (Exception,) as e:
                self.assertTrue(isinstance(e, InvalidConfigurationException))
                self.assertTrue("no_such_validation" in str(e))

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class Test_JSON_ValidationFail(Test_JSON_Validation):
    def test_005_custom_expfail(self):
        try:
            data = self.j_data.copy()
            data["var6"] = 7
            self.set_expectation(
                "custom", validator=CustomDictValidator("var"), exp="N/A", active=True
            )
            try:
                try:
                    self.check_expectations(data=data)
                    self.fail("should have errored 6<>7")
                except (AssertionError,) as e:
                    self.assertTrue("6" in str(e) and "7" in str(e))
            # pragma: no cover pylint: disable=unused-variable
            except (Exception,) as e:
                if cpdb():
                    pdb.set_trace()
                raise

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class PretendHttpValidations:
    cls_validators = [
        ValidationDirective(
            "status_code", validator=DirectValidator("status_code"), exp=200
        ),
        ValidationDirective(
            "content_type", validator=DirectValidator("content_type"), exp="html"
        ),
    ]


class Test_Direct_Validation(
    PretendHttpValidations, BaseForJson, LazyMixinBasic, unittest.TestCase
):
    """ we can also check expectations directly against the TestCase
        For example, suppose your unit test class tracked attributes
        after getting a response through `requests`


    """

    cls_validators = [
        # this overrides the expectations set before
        ValidationDirective("content_type", exp="json")
    ]

    def test_001_expects(self):
        try:

            self.content_type = "json"
            self.status_code = 200

            self.check_expectations(data=self.j_data)
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_002_fails_on_html(self):
        try:

            self.content_type = "html"
            self.status_code = 200
            try:
                self.check_expectations(data=self.j_data)
            except (AssertionError,) as e:
                self.assertTrue("html" in str(e) and "json" in str(e))
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class Test_Direct_AutoExpectations(
    PretendHttpValidations, BaseForJson, LazyMixinBasic, unittest.TestCase
):

    cls_validators = [
        # this overrides the expectations set before we are now expecting to get them from the data
        ValidationDirective(
            "content_type",
            exp=AutoExp("http_content_type"),
            validator=DictValidator("content_type"),
        ),
        ValidationDirective("status_code", active=False),
        ValidationDirective("title", exp=AutoExp, validator=DictValidator("title")),
    ]

    j_data = dict(
        var1=1,
        var2=dict(var21=21, var22=22, var23=dict(var231=231)),
        var3=3,
        var4=4,
        someval="xyz",
        title="mytitle",
        content_type="not_xml",
    )

    def setUp(self):
        self.j_data = self.j_data.copy()

    http_content_type = "not_xml"

    def test_001_expects(self):
        """
            The 2 AutoExp validators, combined with setting the attributes
            on the unittest.TestCase insstance or class , is really the equivalent of:

            self.set_expectation("title",exp="mytitle")
            self.set_expectation("content_type",exp="json")

        """

        try:

            self.title = "mytitle"

            self.check_expectations(data=self.j_data)
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def test_002_fail(self):
        """
        make sure it errors when appropriate
        """

        try:

            data = self.j_data.copy()
            data["title"] = "mytitle2"

            self.http_content_type = "changed_expectations"
            self.title = "mytitle2"

            try:
                self.check_expectations(data=data)
                self.fail("%s should have failed" % self.http_content_type)
            # pragma: no cover pylint: disable=unused-variable
            except (AssertionError,) as e:
                self.assertTrue(self.http_content_type in str(e))

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


if __name__ == "__main__":

    cpdb = set_cpdb()

    sys.exit(unittest.main())



import os
import sys
import unittest
import unittest.mock as mock

#################################################################
# various helpers to simulate an http server
#################################################################

from lazy_regression_tests.helper_tst import (
    Helper,
    get_mock_env,
    get_fake_response_from_template,
)

from lazy_regression_tests._baseutils import set_cpdb, ppp

verbose = "-v" in sys.argv

di_mock_env = get_mock_env()


#################################################################

# generic lazy functionality
from lazy_regression_tests import (
    LazyMixin,
    ValidationDirective,
    FilterDirective,
    AutoExp,
    RegexRemoveSaver,
)

# specific to HTTP/HTML validations
from lazy_regression_tests.http_validators import (
    HTMLValidationMixin,
    ResponseHTML,
    HtmlFilterManager,
    CSSRemoveFilter,
    CSSValidator,
)


class LazyMixinBasic(LazyMixin):
    """ base Mixin class for the lazy test framework """

    lazy_filename = LazyMixin.get_basename(__name__, __file__, __module__)
    cls_filters = dict(html=HtmlFilterManager())


# 👇  🥦🥦 filter changing contents
filter_variables = [
    FilterDirective(
        "timestamp", filter_=CSSRemoveFilter("span.timestamp", "timestamp", scalar=True)
    ),
    FilterDirective("csrf", filter_=RegexRemoveSaver("csrf_token", "csrf_token")),
]


class Test_Features(Helper, HTMLValidationMixin, LazyMixinBasic, unittest.TestCase):

    cls_filters = dict(html=filter_variables)  # 👈 This is how we add the filters

    name = "Mr. Rabbit"  # picked up by `AutoExp` below

    # 👇 setting up the validations 🥦🥦
    cls_validators = [
        ValidationDirective("title", exp="Your order"),
        ValidationDirective("name", exp=AutoExp, validator=CSSValidator("#name")),
    ]

    @mock.patch.dict(os.environ, di_mock_env)
    def test_it(self):
        try:
            # this comes from `requests`, could be a Django testserver, your code...
            http_response = get_fake_response_from_template(self)

            # if we re-use filters and validations, yes, we use a lot less than
            # 10 lines of code 👇 🍰🍰🍰🍰

            # "adapt" standard http_response by tracking content_type, status_code, headers...
            response = ResponseHTML(http_response)
            # Check validation such as content_type and status code
            self.check_expectations(response=response)
            # Regression test - did we get the same contents as the last time?
            tmp = self.assert_exp(response.content, "html")

        except (Exception,) as e:
            raise

    #################################################################
    # template and configuration to fake an http response
    #################################################################

    line1 = "Item 1"
    line2 = "Item 2"
    line3 = "Item 3"

    # the template used to generate the fake html
    template = """
<title>Your order</title>
<script>
const csrf_token = '{{csrf}}';
</script>
<body>
    Hi&nbsp;<span id="name">{{ name }}</span>&nbsp;!<br/> 
    It is now<span class="timestamp">{{ timestamp }}</span>.<br/>
    Your order is:
    <ul>
        <li class="orderline">{{line1}}</li>
        <li class="orderline">{{line2}}</li>
        <li class="orderline">{{line3}}</li>
    </ul>
</body>
    """


if __name__ == "__main__":

    cpdb = set_cpdb()
    rc = 0
    try:
        rc = unittest.main()
    finally:
        sys.exit(rc)



""" this file provides common functionality and constants to the others 
    
    DO NOT import anything lazy in it other than utils/baseutils otherwise
    there is a strong chance for circular import issues.

"""

import pdb

import sys

verbose = "-v" in sys.argv

undefined = NotImplemented

OPT_DIRECTIVE_SKIP = "skip"
OPT_DIRECTIVE_BASELINE = "baseline"
OPT_DIRECTIVE_NODIFF = "nodiff"

from lazy_regression_tests.utils import (
    InvalidConfigurationException,
    ppp,
    UnavailableLibrary,
)


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb


def format_assertion_error(
    testee: "TestCase", assert_exc: AssertionError, validator=None, name: str = None
):
    """ enhance AssertionError's message, if possible """

    try:
        new_message = f"""

❌❌❌❌❌❌❌

Validation failure @ {name} [{validator}]:

{assert_exc}

❌❌❌❌❌❌❌
"""

        assert_exc.args = (new_message,)

    # pragma: no cover pylint: disable=unused-variable
    except (Exception,) as e:
        if cpdb():
            pdb.set_trace()
        raise assert_exc



import pdb
import sys
import re
import json

# pylint: disable=unused-import
from lazy_regression_tests._baseutils import debugObject, ppp, Dummy, getpath
from traceback import print_exc as xp

# pylint: enable=unused-import


#######################################################
# Dependencies
#######################################################

# from bs4 import BeautifulSoup as bs

#######################################################
# Typing
#######################################################
from typing import (
    Optional,
    # TYPE_CHECKING,
    Any,
    cast,
)


from lazy_regression_tests.utils import (
    nested_dict_get,
    nested_dict_pop,
    first,
    fill_template,
    ppp,
    RescueDict,
    Subber,
    DictFormatter,
    InvalidConfigurationException,
)


verbose = "-v" in sys.argv

undefined = NotImplemented

# https://stackoverflow.com/questions/6102019/type-of-compiled-regex-object-in-python
# yeah, it is what it is, so be it...
type_regex_hack = re.compile("").__class__


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb

#######################################################
# Filters
#######################################################


class DataMatcher(object):

    name: str = None
    scalar = False
    selector: str = None

    def __repr__(self):
        try:
            return f"{self.__class__.__name__:30.30} selector:{self.selector} name:{self.name}"
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,):
            return "%s.%s:%s" % (self.__module__, self.__class__.__name__, self.name)

    def __init__(self, selector, name):
        self.selector = selector
        self.name = name

    def format_filtered_hook(self, value):
        return value

    def add_to_filter_hit(self, tmp, value):
        try:

            value = self.format_filtered_hook(value)

            tmp.add_filtered(self.name, value, self.scalar)
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class RawFilter:
    raw = True


class TextFilter:
    raw = False


class RegexMatcher(TextFilter, DataMatcher):
    def __init__(self, pattern, name, scalar=False, *args):
        self.patre = re.compile(pattern, *args)
        self.name = name
        self.scalar = scalar

    def search(self, *args, **kwds):
        return self.patre.search(*args, **kwds)

    def __getattr__(self, attrname):
        return getattr(self.patre, attrname)

    if sys.version_info <= (3, 7):

        def __deepcopy__(self, *args, **kwargs):
            """ bit of a cheat here, because before Python 3.7 regex can't be deepcopied
                shouldn't be too much of a problem as the RegexMatchers aren't meant to be modified once
                created, but it still means that instances are the same across all copied Regex-based filters
            """

            return self


class DictFilter(RawFilter, DataMatcher):

    scalar = True
    dict_ = "?"

    def __repr__(self):
        return "%s[%s]=%s" % (self.__class__.__name__, self.name, self.dict_)

    def __init__(self, dict_: dict, name: str):
        assert isinstance(dict_, dict)
        self.dict_ = dict_
        self.name = name
        self.formatter = DictFormatter(di_default_formatter=dict_)

    def filter(self, options, tmp, data, callback=None):
        try:
            keep = {}
            data = self.formatter.process(data, keep=keep, in_place=False)
            self.add_to_filter_hit(tmp, keep)
            return data

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class RegexFilter(RegexMatcher):
    def process_line(self, hit, line):
        try:
            raise NotImplementedError(
                "%s.process_line(%s).  need to implement on subclass" % (self, locals())
            )
        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise

    def filter(self, options, tmp, data, callback=None):

        li = []

        try:
            line_out = []
            for line in data.split("\n"):
                hit = self.patre.search(line)
                if hit:
                    line2 = self.process_line(hit, line)
                    line_out.append(line2)
                    self.add_to_filter_hit(tmp, line)
                else:
                    line_out.append(line)

            if callback:
                callback(self.name, data, li)

            return "\n".join(line_out)
        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise

    __call__ = filter


class RegexSubstitHardcoded(RegexFilter):
    """allows for replacement of the line with different contents

       can't use a re.sub directly because the Filter won't know if 
       it's just a match filter or a match & substitution
    """

    def process_line(self, hit, line):
        try:

            for from_, to_ in self.subs:
                if isinstance(from_, str):
                    line = line.replace(from_, to_)
                elif hasattr(from_, "sub"):
                    line = from_.sub(to_, line)
                else:
                    raise NotImplementedError(
                        f"{self}.process_line(from_:{from_}, to_:{to_})"
                    )

            return line

        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise

    def __init__(self, pattern, name, subs, scalar=False, *args):
        super(RegexSubstitHardcoded, self).__init__(pattern, name, scalar)
        self.subs = subs


class RegexRemoveSaver(RegexMatcher):
    """this will remove the matching line but also save it"""

    def filter(self, options, tmp, data, callback=None):

        li = []

        try:
            line_out = []
            lines = data.split("\n")
            for ix, line in enumerate(lines):
                hit = self.patre.search(line)
                if hit:
                    li.append(line)
                    self.add_to_filter_hit(tmp, line)

                else:
                    line_out.append(line)

            if callback:
                callback(self.name, data, li)

            return "\n".join(line_out)
        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise

    __call__ = filter


#######################################################
# Filter management
#######################################################


class FilterDirective:
    """ Tracks combinations of Filters, active flags, callbacks by name
        
    """

    name = filter_ = active = callback = None

    def __repr__(self):

        try:
            return f"{self.__class__.__name__:30.30} active:{self.active} filter:{str(self.filter_):30.30}"

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            pdb.set_trace()
            return "%s.%s" % (self.__class__.__name__, self.name)
            # raise

        return "%s:%s active:%s callback=%s with %s\n" % (
            self.__class__.__name__,
            self.name,
            self.filter_,
            self.active,
            (self.callback.__name__ if self.callback is not None else None),
        )

    def __init__(
        self, name: str, filter_: Any = None, active: bool = None, callback=None
    ):
        try:
            self.name = name

            # swap arguments if keywords not used
            if isinstance(filter_, bool) and active is None:
                active = filter_
                filter_ = None

            assert active is None or isinstance(active, bool)

            assert filter_ is None or isinstance(
                filter_, DataMatcher
            ), "unexpected filter_ type: %s" % (filter_)

            assert callback is None or callable(callback)

            self.raw = getattr(filter_, "raw", False)

            self.filter_ = filter_
            self.active = active if active in (True, False) else bool(filter_)
            self.callback = callback

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def filter(self, *args, **kwargs):
        return self.filter_.filter(*args, **kwargs)

    def copy(self):
        return self.__class__(
            name=self.name,
            filter_=self.filter_,
            active=self.active,
            callback=self.callback,
        )


class FilterManager:
    """ a filter mgr allows tracking of filters by name, via FilterDirectives

        its core method is  `set_filter` 
    """

    def prep(self, tmp, data):
        """default do-nothing implementation"""
        try:
            return data
        # pragma: no cover pylint: disable=unused-variable, broad-except
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def to_text(self, tmp, data):
        """default do-nothing implementation"""
        try:
            return str(data)
        # pragma: no cover pylint: disable=unused-variable, broad-except
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def __repr__(self):

        try:
            filters = sorted(self.filters.keys())
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            filters = self.filters

        sub = dict(
            id_="id=%s" % id(self) if verbose else "",
            name=self.__class__.__name__,
            filters=filters,
        )

        tmpl = "%(name)s[%(id_)s] filters=%(filters)s"  # % (sub, self)

        return fill_template(tmpl, sub, self, RescueDict())

    def __init__(self, *filters_in):
        self.filters = {}
        for filter_ in filters_in:
            self.set_filter(filter_)

    def get_raw_text_filters(self):
        """docstring"""
        try:

            rawfiltermgr = RawFilterManager()
            textfiltermgr = TextFilterManager()
            for name, directive in self.filters.items():
                if directive.active is not True:
                    continue

                filter_ = directive.filter_

                if filter_ is None:
                    raise InvalidConfigurationException(
                        f"Directive.{name} : {directive} is active. without a filter"
                    )

                if isinstance(filter_, RawFilter):
                    rawfiltermgr.set_filter(directive)
                elif isinstance(filter_, TextFilter):
                    textfiltermgr.set_filter(directive)
                else:
                    raise InvalidConfigurationException(
                        "Directive.%s uses an unknown FilterType.  Filters need be either RawFilter or TextFilter subclasses"
                        % (directive)
                    )

            return rawfiltermgr, textfiltermgr

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def filter(self, options, tmp, data):
        """applies its filters to incoming data"""
        try:
            callbacks = getattr(options, "reg_callbacks", {})

            for name, directive in self.filters.items():

                if not directive.active:
                    continue

                callback = callbacks.get(directive.name)
                data = directive.filter(options, tmp, data, callback)
            return data

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def set_filter(self, name, filter_=None, active=None, callback=None):
        """
        allows successive adjustments of the filters

        you can add FilterDirectives, FilterMgr, or specify everything via
        this method

        Note the importance of using directive.copy:  by intent FilterDirectives and
        FilterMgrs can be shared accross a test suite, so changing any element should
        only happen on copies.

        """

        if breakpoints("set_filter", {"name": name}):  # pragma: no cover
            pdb.set_trace()

        try:

            if isinstance(name, FilterDirective):
                directive = name.copy()
                self += directive

            elif isinstance(name, FilterManager):

                # hmmm, wonder if we could override the FilterManager class here....

                for directive in name.filters.values():
                    self += directive.copy()
            else:
                directive = FilterDirective(
                    name=name, filter_=filter_, active=active, callback=callback
                )
                self += directive

            return self

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def __iadd__(self, directive):
        """ adds a FilterDirective to `self.filters`
            if a directive by that name already existed, it is updated.
        """

        try:

            existing = self.filters.get(directive.name)
            if existing:

                if directive.active in (False, True):
                    existing.active = directive.active

                if directive.filter_:
                    existing.filter_ = directive.filter_

                if directive.callback:
                    existing.callback = directive.callback

            else:
                self.filters[directive.name] = directive

            return self

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def copy(self):
        newinst = self.__class__()
        for name, filter_ in self.filters.items():
            newinst.filters[name] = filter_.copy()
        return newinst

    def debug(self, testee):

        try:
            padder = "⚙️" * 6 + "  lazy-tests configuration  " + "⚙️" * 6
            print(f"\n{padder}\n{self}  filters:")

            for key in self.filters:
                value = self.filters[key]
                print(
                    f"  {key:20.20}: active:{str(value.active):5.5} filter:{str(value.filter_):30.30}"
                )

            li_revmro = list(reversed(testee.__class__.mro()))

            li = [
                (cls, getattr(cls, "cls_filters", {}).get("html")) for cls in li_revmro
            ]

            li = [tu_ for tu_ in li if tu_[1]]

            print(
                "\n\n ⚙️class-level inheritance:\n%s\n"
                % "\n".join([tu[0].__name__ for tu in li])
            )

            actual = None
            for cls, filt_ in li:
                oi = filt_.filters
                # print(f"\n{cls.__name__:30.30} {filt_} filters.id:{id(filt_.filters)}")

                for key, actual in oi.items():
                    filter_ = getattr(actual, "filter_", actual)
                    print(
                        f"  {key:20.20} : active:{actual.active} {str(filter_):40.40} id actual:{id(actual)}"
                    )

            print(f"\n{padder}\n")

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if 1 or cpdb():
                pdb.set_trace()
            raise


class RawFilterManager(FilterManager):
    filter_cls = RawFilter


class TextFilterManager(FilterManager):
    filter_cls = TextFilter


#######################################################
# Filters
#######################################################


def build_filters_for_class(cls, li_ancestor_filter):
    """build filters for class, before instance overrides"""
    try:

        classname = cls.__name__

        # breakpoints dont work for now as the class defs are an import time execution
        # not a call time execution
        if breakpoints(
            "build_filters_for_class", dict(classname=classname)
        ):  # pragma: no cover
            pdb.set_trace()

        # remember, the mro for a class includes the class itself
        if not li_ancestor_filter:
            return {}

        if len(li_ancestor_filter) == 0:
            raise NotImplementedError("%s.build_filters_for_class(%s)" % (cls, {}))

        if len(li_ancestor_filter) >= 1:

            # first, build a set of all extensions
            s_extension = set()
            for di in li_ancestor_filter:
                s_extension |= set(di.keys())

            finals = getattr(cls, "cls_filters", {})

            ancestors = li_ancestor_filter

            res = {}

            for extension in s_extension:
                li = [di.get(extension) for di in ancestors]
                li = [di for di in li if di]

                mgr_cls = None
                mgr = finals.get(extension)

                if not isinstance(mgr, FilterManager):

                    mgr = None
                    for cand in li:
                        if isinstance(cand, FilterManager):
                            mgr = cand

                    mgr_cls = mgr.__class__ if mgr else FilterManager

                else:
                    mgr_cls = mgr.__class__

                mgr_cls = mgr_cls or FilterManager

                newfilter = mgr_cls()
                for directive3 in li:
                    if not isinstance(directive3, list):
                        directive2 = [directive3]
                    else:
                        directive2 = directive3

                    for directive in directive2:
                        newfilter.set_filter(directive)

                res[extension] = newfilter

            return res
            # pdb.set_trace()

        else:
            return last_

    # pragma: no cover pylint: disable=unused-variable
    except (Exception,) as e:
        if 1 or rpdb() or cpdb():
            pdb.set_trace()
        raise


class JsonFilterManager(FilterManager):
    """Filter Manager for JSON/dict content"""

    def prep(self, tmp, data):
        try:
            if isinstance(data, dict):
                return data
            elif isinstance(data, str):
                return json.loads(data)
            else:
                raise NotImplementedError(
                    "%s.prep:unsupported data type:%s" % (self, type(data))
                )

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def to_text(self, tmp, data):
        try:
            return json.dumps(data, sort_keys=True, indent=4)
        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise



import re
import os
import difflib
from collections import deque, namedtuple
from functools import partial


import pdb

from lazy_regression_tests._baseutils import (
    set_cpdb,
    set_rpdb,
    ppp,
    debugObject,
    fill_template,
    Subber,
    RescueDict,
    nested_dict_get,
    nested_dict_pop,
    first,
    getpath,
    DictFormatter,
    UnavailableLibrary as BaseUnavailableLibrary,
)


from traceback import print_exc as xp  # pylint: disable=unused-import

import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

undefined = NotImplemented


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb


class InvalidConfigurationException(ValueError):
    """ indicates that some configuration is missing or invalid """

    def __init__(self, msg, *args, **kwargs):

        new_message = f"""
❌⚙️❌⚙️❌⚙️❌

{msg}

❌⚙️❌⚙️❌⚙️❌
"""
        super(InvalidConfigurationException, self).__init__(
            new_message, *args, **kwargs
        )


class UnavailableLibrary(BaseUnavailableLibrary):
    """ exception with message detailing what was not installed 
        with emoji banner 
    """

    exception_cls = InvalidConfigurationException

    def __getattr__(self, *args, **kwargs):
        """ throw errors on any access """
        raise self.exception_cls(self)

    # and reroute call
    __call__ = __getattr__



""" support jsonschema based validations 

see https://pypi.org/project/jsonschema/

This is still somewhat experimental

"""

#################################################################
# debugging support
#################################################################
# pylint: disable=unused-import

import pdb

from traceback import print_exc as xp
from pprint import pprint as xpp

from lazy_regression_tests._baseutils import first, ppp, getpath


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb

#######################################################
# and Typing...
#######################################################
from typing import (
    Optional,
    # TYPE_CHECKING,
    Any,
    cast,
)

#######################################################


# pylint: enable=unused-import


#######################################################
# Dependencies
#######################################################


try:
    from jsonschema import validate
    import jsonschema.exceptions
# pragma: no cover pylint: disable=unused-variable
except (ImportError,) as e:

    class Foo:
        """dummy class"""

    # this will throw an InvalidConfigurationError on any access to ydump
    # telling you to install yaml
    from .common import UnavailableLibrary

    ValidationError = Exception

    jsonschema = Foo
    jsonschema.exceptions = Foo
    jsonschema.exceptions.ValidationError = NotImplementedError

    validate = UnavailableLibrary(name=Foo.__module__, missing="jsonschema")


undefined = NotImplemented


from lazy_regression_tests.validators import DictValidator


class JsonSchemaValidator(DictValidator):
    """
    This is still somewhat experimental
   Implements jsonchema based validations 
    """

    def __init__(self, sourcename: str = "data", selector: str = None, cargo=None):
        """ init """
        try:
            self.sourcename = sourcename
            self.cargo = cargo
            assert isinstance(sourcename, (str, None.__class__))
            self.selector = selector
        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise

    def get_value(self, source_):

        if self.selector is None:
            return source_
        else:
            return getpath(source_, self.selector)

    def check(self, name: str, testee: "LazyMixin", exp: dict, sources: dict):
        """ validate using the schema """

        try:

            igot = None
            source_ = self.get_source(testee, **sources)
            got = self.get_value(source_)

            if not isinstance(got, list):
                got = [got]

            for igot in got:
                try:
                    validate(instance=igot, schema=exp)
                # pragma: no cover pylint: disable=unused-variable
                except (jsonschema.exceptions.ValidationError,) as e:
                    testee.fail(e)

        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                ppp(exp, "schema")
                ppp(igot, "igot")

                pdb.set_trace()
            raise



# -*- coding: utf-8 -*-
"""
test lazy-regression-tests
"""
import sys
import os
import re

import unittest

import pdb

try:
    import unittest.mock as mock
except (ImportError,) as ei:
    import mock  # python 2?


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


from lazy_regression_tests._baseutils import set_cpdb, ppp

rpdb = breakpoints = cpdb

from lazy_regression_tests.http_validators import HTMLValidationMixin, ResponseHTML


from lazy_regression_tests import LazyMixin

from lazy_regression_tests import ValidationDirective, AutoExp


from lazy_regression_tests.http_validators import CSSValidator

#################################################################
# Pre-defined lazy-testing
#################################################################

from lazy_regression_tests.filters import (
    # a filter class that operates on dictionaries
    DictFilter,
    FilterDirective,
    # works on data, once it has been converted to
    # text and before it gets saved to file or compared
    RegexRemoveSaver,
)

from lazy_regression_tests.http_validators import (
    # provides save-to-file, load-from-file functionality
    # and calls BeautifulSoup to get a DOM
    HtmlFilterManager,
    # operates on a BeautifulSoup DOM, using standard CSS
    # selectors
    CSSRemoveFilter,
)


#################################################################
# These aren't really objects-under-test, they're used to
# simulate a Django-style HTMLResponse.
#################################################################

from lazy_regression_tests.helper_tst import (
    Helper,
    get_mock_env,
    get_fake_response_from_template,
)

get_fake_html_response = get_fake_response_from_template

di_mock_env = get_mock_env()

verbose = "-v" in sys.argv


DO_SKIPPED_TESTS = False

# pylint: disable=no-member
class LazyMixinBasic(LazyMixin):
    """ base Mixin class for the lazy test framework """

    # 👇 ⚙️ tracks where expectation and received files are saved
    lazy_filename = LazyMixin.get_basename(__name__, __file__, __module__)

    extension = "html"

    # 👇 ⚙️ Tells the framework what extensions/content to expect
    cls_filters = dict(html=HtmlFilterManager())

    @mock.patch.dict(os.environ, di_mock_env)
    def test_it(self):
        """ fetch data, run validations, regression test """
        try:

            # this is a pretend server, could have been the Django test server for example
            # could be `requests` against a real site
            # anything returning an HTML response would do
            http_response = get_fake_html_response(self)

            # 👇 lazy-testing, in 3 lines of code 👇

            # ResponseHTML "adapts" the standard http_response by tracking attributes
            # like content_type, status_code, headers...
            response = ResponseHTML(http_response)

            # Check validation such as content_type and status code
            self.check_expectations(response=response)

            # Regression test - did we get the same contents as the last time?
            tmp = self.assert_exp(response.content, "html")

            # 👆 lazy-testing 👆

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


filter_variables = [
    FilterDirective(
        "timestamp", filter_=CSSRemoveFilter("span.timestamp", "timestamp", scalar=True)
    ),
    FilterDirective("csrf", filter_=RegexRemoveSaver("csrf_token", "csrf_token")),
]


#################################################################
# PYLINT:
#################################################################
# With this style of Mixin coding where filters and validators
# are composed onto the unittest.TestCase subclasses
# there WILL be many ancestors.  That's the point.
#################################################################

# pylint: disable=too-many-ancestors

#################################################################
# These however are just disabled to make the sample code
# easier to paste into documentation.
#################################################################

# pylint: disable=attribute-defined-outside-init
# pylint: disable=import-outside-toplevel

#################################################################


#  ⚙️   This enables the lazy-test framework     👇
class Test_Features(Helper, HTMLValidationMixin, LazyMixinBasic, unittest.TestCase):
    """ this is the test we are running here """

    cls_filters = dict(html=filter_variables)  # 👈 This is how we add the filters

    name = "Mr. Rabbit"
    line1 = "Item 1"
    line2 = "Item 2"
    line3 = "Item 3"

    # 👇 setting up the validations
    cls_validators = [
        ValidationDirective("title", exp="Your order"),
        ValidationDirective("name", exp=AutoExp, validator=CSSValidator("#name")),
    ]

    # the template used to generate the fake html
    template = """
<title>Your order</title>
<script>
const csrf_token = '{{csrf}}';
</script>
<body>
    Hi&nbsp;<span id="name">{{ name }}</span>&nbsp;!<br/> 
    It is now<span class="timestamp">{{ timestamp }}</span>.<br/>
    Your order is:
    <ul>
        <li class="orderline">{{line1}}</li>
        <li class="orderline">{{line2}}</li>
        <li class="orderline">{{line3}}</li>
    </ul>
</body>
    """


class Test_Features_Regex(Test_Features):
    """ This should fail """

    cls_validators = [  # 👇                      #👇
        ValidationDirective(
            "item", exp=re.compile("^Item"), validator=CSSValidator("li.orderline")
        )
    ]

    name = "Mr. Rabbit"
    line1 = "Item 1"
    line2 = "Item 2"
    line3 = "Bad line 3"

    def test_it(self):
        try:
            Test_Features.test_it(self)
        # pragma: no cover pylint: disable=unused-variable
        except (AssertionError,) as e:
            self.assertTrue("Bad" in str(e))


def check_lineitems(testee: "unittest.TestCase", got, validator: "Validator"):
    """
    `got` will be a list of strings here 
    👉keeping this as stand alone function, rather a method of your TestCase
    means it can be used anywhere.  It still behaves just like a test method
    """
    try:
        for igot in got:
            if not igot.endswith("3"):
                testee.assertTrue(igot.startswith("Item"))
    # pragma: no cover pylint: disable=unused-variable
    except (Exception,) as e:
        if cpdb():
            pdb.set_trace()
        raise


class Test_Features_CustomLineValidation(Test_Features_Regex):
    """ This should pass, we are re-using the CSSValidation lookup for `item`"""

    #   👇
    cls_validators = [ValidationDirective("item", exp=check_lineitems)]  #   👇

    name = "Mr. Rabbit"
    line1 = "Item 1"
    line2 = "Item 2"
    line3 = "Bad line 3"

    def test_it(self):
        Test_Features.test_it(self)


class Test_Turning_ThingsOff(Test_Features):
    """ we don't have a title or a greeting anymore
        and we don't need to filter out the timestamp either
    """

    template = """
<body>
    It is now<span class="timestamp">fake, fixed, timestamp</span>.<br/>
</body>
"""

    def setUp(self):
        # 👇 turn these off to avoid validation errors
        self.set_expectation("title", active=False)
        self.set_expectation("name", active=False)
        self.filters["html"].set_filter("timestamp", active=False)  # 👈 keep it


class Test_404(Test_Features):
    @mock.patch.dict(os.environ, di_mock_env)
    def test_it(self):
        """ if you have a lot of config to access a particular URL
        you don't want to duplicate it all over again to simulate a 
        404.  Yet, your 404 page may not at all have the same contents
        or title.
        """
        try:

            http_response = get_fake_html_response(self)
            response = ResponseHTML(http_response)

            # fake a 404
            response.content = "<div>404, to you, buddy!</div>"
            response.status_code = 404

            # This handles the 404, however, the title and name would error out
            self.set_expectation("status_code", 404)

            # 👇 turn off what you don't need
            constants_keep_on_404 = ["status_code", "content_type"]
            self.check_expectations(
                response=response, lazy_skip_except=constants_keep_on_404
            )

            tmp = self.assert_exp(response.content, "html")

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


#################################################################
# Test JSON and YAML
#################################################################


class Test_JSON_Too(LazyMixinBasic, unittest.TestCase):
    """ just connect it to the appropriate filter manager for 
    the extension type
    """

    from lazy_regression_tests.filters import JsonFilterManager

    cls_filters = dict(json=JsonFilterManager())
    extension = "json"

    exp_fail = "told"

    @mock.patch.dict(os.environ, di_mock_env)
    def test_it(self):
        """ simulate data changes """
        try:
            data = dict(var1="the_same", var2="will_change")
            tmp = self.assert_exp(data, self.extension)
            data.update(var2="told you so")

            if self.exp_fail:
                try:
                    tmp = self.assert_exp(data, self.extension)
                    self.fail(f"should failed with {self.exp_fail}")
                except (AssertionError,) as e:
                    self.assertTrue(
                        str(self.exp_fail) in str(e),
                        f"should have failed with {self.fail} in {e}",
                    )
            else:
                tmp = self.assert_exp(data, self.extension)

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class Test_JSON_Filter(Test_JSON_Too):
    """ let's fix the above error by filtering out the changing key"""

    cls_filters = dict(
        json=FilterDirective(
            "changing",
            # DictFilter work by looking for matching keys in the target dictionary.
            # and then call their value if it's a callable.
            # value=None is the default behavior and just deletes the key in the target
            DictFilter(dict(var2=None), "changing"),
        )
    )


class Test_YAML(Test_JSON_Too):
    """ hey, most of the work was done by the JSON guys already
    """

    from lazy_regression_tests.yaml_validators import YAMLFilter

    extension = "yaml"
    cls_filters = dict(yaml=YAMLFilter())


#################################################################
# Test arbitrary object graphs
#################################################################


class Subvar:
    "dummy class"

    def __init__(self, value):
        """ init """
        self.value = value


class SomethingToTest:
    "dummy class"

    def __init__(self):
        """ init """
        self.var1 = 11
        self.var2 = 12
        self.var4 = dict(FF="Fantastic")


class Test_YAML_Graphs(Test_YAML):
    """ we can look at arbitrary serializable objects 
        For example, if our Order object is always
        intended to have order.customer populated this
        would catch a case where a particular testcase did
        originally have this result, but a latter run
        did not.
    """

    @unittest.expectedFailure
    @mock.patch.dict(os.environ, di_mock_env)
    def test_down_the_rabbit_hole(self):
        """ simulate a changed object graph """
        try:

            from yaml import dump as ydump, safe_load as yload

            somethingtotest = SomethingToTest()
            somethingtotest.var3 = Subvar("3")

            yaml_ = ydump(somethingtotest)

            # probably not a good idea with untrusted data
            data = yload(yaml_)

            somevar = "somevalue"

            self.assert_exp(data, self.extension)
            somethingtotest.added_this = dict(somevar=somevar)
            somethingtotest.var3.value = "3++"

            yaml_ = ydump(somethingtotest)

            # probably not a good idea with untrusted data
            data = yload(yaml_)

            self.assert_exp(data, self.extension)

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


if __name__ == "__main__":

    cpdb = set_cpdb()
    rc = 0
    try:
        rc = unittest.main()
    finally:
        sys.exit(rc)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["timeout-decorator"]

setup_requirements = []

test_requirements = ["requests", "responses", "Jinja2", "jsonschema", "beautifulsoup4"]

setup(
    version="0.5.3",
    author="JL Peyret",
    author_email="jpeyret@gmail.com",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        # 2.x could come back, valuable for migrations,
        # but concentrating on 3.x for now
        # "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        # No 3.5, uses f-strings:
        #'Programming Language :: Python :: 3.5',
        "Programming Language :: Python :: 3.6",
        # should work, not tested:
        "Programming Language :: Python :: 3.7",
        # should work, not tested:
        "Programming Language :: Python :: 3.8",
    ],
    description="Easier automatic validation and regression testing",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="lazy_regression_tests",
    long_description_content_type="text/markdown",
    name="lazy_regression_tests",
    packages=find_packages(
        include=["lazy_regression_tests", "lazy_regression_tests.*"]
    ),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/jpeyret/lazy-regression-tests",
    zip_safe=False,
)



# -"""Top-level package for lazy regression tests."""

__author__ = """JL Peyret"""
__email__ = "jpeyret@gmail.com"
__version__ = "0.5.3"


from .core import LazyMixin, ValidationManager

from .filters import FilterManager, FilterDirective, RegexRemoveSaver

from .validators import (
    NamedTesteeAttributeValidator,
    FullyQualifiedNamesValidator,
    MixinExpInGot,
    ValidationDirective,
    DictValidator,
    DirectValidator,
    AutoExp,
)



# -*- coding: utf-8 -*-
"""
test http behavior - requires BeautifulSoup, requests and responses (a mock for requests)
"""

import sys
import os
import pdb

import unittest

try:
    import unittest.mock as mock
except (ImportError,) as ei:
    import mock  # python 2?

import datetime
import re


# pylint: disable=unused-import
#######################################################
# Typing
#######################################################
from typing import (
    Optional,
    # TYPE_CHECKING,
    Any,
    cast,
)

#######################################################
# pylint: enable=unused-import


verbose = "-v" in sys.argv


# pylint: disable=attribute-defined-outside-init

# pylint: disable=missing-function-docstring,missing-class-docstring  #🧨🧨🧨🧨 turn this back on later 🧨🧨🧨🧨

import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)
from traceback import print_exc as xp  # pylint: disable=unused-import

from lazy_regression_tests._baseutils import RescueDict, set_cpdb

from lazy_regression_tests import ValidationDirective, AutoExp


from lazy_regression_tests.filters import FilterDirective, RegexRemoveSaver


from lazy_regression_tests.core import OPT_DIRECTIVE_BASELINE


rescuedict = RescueDict()


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb


from lazy_regression_tests import LazyMixin

##########################################################
# tests
##########################################################


lzrt_default_t_basename = "%(filename)s %(classname)s %(_testMethodName)s %(lazy_basename_extras)s %(suffix)s %(extension)s"


from lazy_regression_tests.helper_tst import (
    get_mock_env,
    get_fake_response_from_template,
    HelperHTML,
    CheckitMixin,
)

di_mock_env = get_mock_env()

di_mock_env_baseline = di_mock_env.copy()


di_mock_env_baseline.update(lzrt_directive=OPT_DIRECTIVE_BASELINE)


di_mock_env_no_extras = dict(
    [(k, v.replace("/%(lazy_dirname_extras)s", "")) for k, v in di_mock_env.items()]
)

module_ = "builtins"
module_ = module_ if module_ in sys.modules else "__builtin__"

from lazy_regression_tests.http_validators import (
    HtmlFilterManager,
    HTMLValidationMixin,
    ResponseHTML,
    CSSValidator,
    CSSRemoveFilter,
)


# pylint: disable=no-member
class LazyMixinBasic(LazyMixin, unittest.TestCase):

    """
    basic support functionality for the rest of the tests
    similar in fact to how LazyMixin is normally used.

    """

    cls_filters = dict(html=HtmlFilterManager())

    lazy_filename = LazyMixin.get_basename(__name__, __file__, __module__)

    tmp_formatted_data = None

    extension = "txt"

    def setUp(self):
        """not sure why needed, but need to reset the environment
            to make sure mocked values are taken into account
           This is probably due to some too aggressive optimization,
           i.e. we-dont-expect-env-variables to change during a test run
        """

        self.lazy_environ.acquired = False

    def seed(self, exp, extension=None, suffix=""):
        try:
            try:

                extension = extension or self.extension

                self.assert_exp(exp, extension=extension, suffix=suffix)

                fnp_exp = self.lazytemp.fnp_exp
                self.lazytemp = None
            # pragma: no cover pylint: disable=unused-variable
            except (AssertionError,) as e:
                fnp_exp = self.lazytemp.fnp_exp
                self.lazytemp = None
            return fnp_exp

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    def get_data(self, seed: dict = None):

        seed = seed or {}

        res = getattr(self, "data", {})
        res.update(**seed)

        return res

    def get(self, url=None, data: dict = None):
        try:

            data = data or {}

            self.data = data

            response = get_fake_response_from_template(testee=self, data=data, url=url)
            return ResponseHTML(response)

        # pragma: no cover pylint: disable=unused-variable
        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


# pylint: disable=too-many-ancestors


class HTMLBase(HTMLValidationMixin, LazyMixinBasic):
    cls_validators = [ValidationDirective("title", active=False)]
    cls_filters = dict(
        html=FilterDirective(
            "timestamp",
            filter_=CSSRemoveFilter("span.timestamp", "timestamp", scalar=True),
        )
    )


class Test_Basic(HTMLBase):

    template = """
        <html><body>Hi <span id="name" >{{ name }}</span>! It is now <span class="timestamp">{{ now }}</span>.</body></html>
    """

    name = "Mr. Rabbit"

    @mock.patch.dict(os.environ, di_mock_env)
    def test_001_basic(self):
        """get response"""
        try:

            data = dict(name=self.name, now=datetime.datetime.now())

            response = self.get(data=data)

            self.check_expectations(response=response)
            tmp = self.assert_exp(response.content, "html")

            data["now"] = "a bit later"
            response = self.get(data=data)

            tmp = self.assert_exp(response.content, "html")

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class Test_Validation(Test_Basic):

    template = """
        <html><head><title>Greetings<title></head>
        <body>Hi <span id="name" >{{ name }}</span>! It is now 
        <span class="timestamp">{{ now }}</span>.
        </body></html>
    """

    cls_validators = [
        ValidationDirective(
            "name", exp=AutoExp, validator=CSSValidator("#name", "name")
        ),
        # we have a title now
        ValidationDirective("title", active=True, exp=AutoExp),
    ]
    name = "Pham Nuwen"
    title = "Greetings"


class HTMLCheck(HTMLBase):

    fail_on_validate = None
    fail_on_change = None

    extension = "html"

    def get_data(self, seed: dict = None):

        seed = seed or {}

        data = getattr(self, "data", {}) or dict(
            name=self.name, now=datetime.datetime.now()
        )
        res = data.copy()
        res.update(**seed)
        return res

    def get_response(self):
        return self.get(data=self.get_data())

    @mock.patch.dict(os.environ, di_mock_env)
    def test_it(self):
        """get response"""
        try:

            if self.__class__ in (HTMLCheck,):
                return

            response = self.get_response()

            if self.fail_on_validate:
                try:
                    self.check_expectations(response=response)
                # pragma: no cover pylint: disable=unused-variable
                except (AssertionError,) as e:
                    self.assertTrue(str(self.fail_on_validate) in str(e))
                else:
                    self.fail(
                        "should have failed validation with %s"
                        % (self.fail_on_validate)
                    )

            else:
                self.check_expectations(response=response)

            if self.fail_on_change:
                try:
                    tmp = self.assert_exp(response.content, self.extension)
                # pragma: no cover pylint: disable=unused-variable
                except (AssertionError,) as e:
                    self.assertTrue(str(self.fail_on_change) in str(e))
                else:
                    self.fail(
                        "should have failed change checks with %s"
                        % (self.fail_on_change)
                    )

            else:
                tmp = self.assert_exp(response.content, self.extension)

            tmp = self.assert_exp(response.content, "html")

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


class Test_ValidationNonScalar(HTMLCheck):
    """
    each <li> is going to be checked against a regex
    """

    name = "Pham"

    template = """
        <html><head><title>Greetings<title></head>
        <body>Hi <span id="name" >{{ name }}</span>! It is now 
        <span class="timestamp">{{ now }}</span>.
        <ul>
            <li>Item 1</li>
            <li>Item 2</li>
            <li>Item 3</li>
        </ul>
        </body></html>
    """

    cls_validators = [
        #
        ValidationDirective(
            "list", exp=re.compile("Item"), validator=CSSValidator("li")
        )
    ]


class Test_ValidationNonScalarFail(Test_ValidationNonScalar):
    """
    the regex will fail
    """

    fail_on_validate = "Not Item"

    cls_validators = [ValidationDirective("list", exp=re.compile("Not Item"))]


class Test_ValidationFailLine3(Test_ValidationNonScalar):

    cls_validators = [
        #
        ValidationDirective(
            "list", exp=re.compile("Item"), validator=CSSValidator("li"), active=True
        )
    ]

    name = "Ezr Vinh"
    line3 = "Line3"

    data = dict(line1="Item 11", line2="Item 12", line3=line3, name=name, now="today")

    fail_on_validate = "Line3"

    template = """
        <html><head><title>Greetings</title></head>
        <body>Hi <span id="name" >{{ name }}</span>! It is now 
        <span class="timestamp">{{ now }}</span>.
        <ul>
            <li>{{line1}}</li>
            <li>{{line2}}</li>
            <li>{{line3}}</li>
        </ul>
        </body></html>
    """


class Helper2(HTMLValidationMixin, HelperHTML, CheckitMixin):
    """ just some boiler plate code to assist in testing 
        No Lazy-related features on bases yet
    """

    extension = "html"


class Test_Features(Helper2, LazyMixinBasic):

    name = "Mr. Rabbit"
    line1 = "Item 1"
    line2 = "Item 2"
    line3 = "Item 3"

    template = """
<title>Greetings</title>
<script>
const csrf_token = '{{csrf}}';
</script>
<body>
    Hi&nbsp;<span id="name">{{ name }}</span>&nbsp;! 
    It is now<span class="timestamp">{{ timestamp }}</span>.
    Your order is:
    <ul>
        <li>{{line1}}</li>
        <li>{{line2}}</li>
        <li>{{line3}}</li>
    </ul>
</body>
    """

    cls_validators = [
        ValidationDirective("title", exp="Greetings"),
        ValidationDirective("name", exp=AutoExp, validator=CSSValidator("#name")),
    ]

    cls_filters = dict(
        html=[
            FilterDirective(
                "timestamp",
                filter_=CSSRemoveFilter("span.timestamp", "timestamp", scalar=True),
            ),
            FilterDirective(
                "csrf", filter_=RegexRemoveSaver("csrf_token", "csrf_token")
            ),
        ]
    )


class Test_Validation_Fail(HTMLBase):

    cls_validators = [
        ValidationDirective(
            "name", exp=AutoExp, validator=CSSValidator("#name", "name")
        ),
        # we have a title now
        ValidationDirective("title", active=True, exp=AutoExp),
    ]

    template = """
        <html><head><title>Greetings<title></head><body>Hi <span id="name" >{{ name }}?</span>! It is now <span class="timestamp">{{ now }}</span>.</body></html>
    """

    title = "Greetings"
    name = "Robert Gu"

    @mock.patch.dict(os.environ, di_mock_env)
    def test_001_fail_name(self):
        """get response"""
        try:

            data = dict(name="Woodcarver", now=datetime.datetime.now())

            response = self.get(data=data)

            try:
                self.check_expectations(response=response)
                self.fail("should have failed name validation")
            except (AssertionError,) as e:
                for text in ["name", "Gu", "Woodcarver"]:
                    self.assertTrue(text in str(e))

            tmp = self.assert_exp(response.content, "html")

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_mock_env)
    def test_002_fail_title(self):
        """get response"""
        try:

            data = dict(name=self.name, now=datetime.datetime.now())

            self.template = self.template.replace("Greetings", "Bonjour")
            response = self.get(data=data)

            try:
                self.check_expectations(response=response)
                self.fail("should have failed title validation")
            except (AssertionError,) as e:
                for text in ["title"]:
                    self.assertTrue(text in str(e))

            tmp = self.assert_exp(response.content, "html")

        except (Exception,) as e:
            if cpdb():
                pdb.set_trace()
            raise


if __name__ == "__main__":

    cpdb = set_cpdb()

    sys.exit(unittest.main())



#################################################################
# YAML Filter Manager
#################################################################

import pdb
from traceback import print_exc as xp

try:
    from yaml import dump as ydump

# pragma: no cover pylint: disable=unused-variable
except (ImportError,) as e:

    class Foo:
        pass

    # this will throw an InvalidConfigurationError on any access to ydump
    # telling you to install yaml
    from .common import UnavailableLibrary

    ydump = UnavailableLibrary(name=Foo.__module__, missing="yaml")


from .filters import FilterManager


def cpdb(*args, **kwargs):
    "disabled conditional breakpoints - does nothing until activated by set_cpdb/rpdb/breakpoint3"


rpdb = breakpoints = cpdb


class YAMLFilter(FilterManager):
    """ note dont try to `prep(data)` on incoming yaml 
        the reason is that a dump in yaml will not safe_load back
    """

    def __repr__(self):
        return "%s" % (self.__class__.__name__)

    def to_text(self, tmp, data):
        try:

            if isinstance(data, str):
                raise NotImplementedError("%s.to_text(data is a string)" % (self))

            return ydump(data, default_flow_style=False)

        except (
            Exception,
        ) as e:  # pragma: no cover pylint: disable=unused-variable, broad-except
            if cpdb():
                pdb.set_trace()
            raise
